from poisson_approval.constants.constants import *
from poisson_approval.random_factories.RandSimplexUniform import RandSimplexUniform
from poisson_approval.utils.Util import initialize_random_seeds
from poisson_approval.profiles.ProfileTwelve import ProfileTwelve


class RandProfileTwelveUniform(RandSimplexUniform):
    """A random factory of twelve-type profiles (:class:`ProfileTwelve`) following the uniform distribution.

    Parameters
    ----------
    types : iterable, optional
        These types will have a variable share. They can be twelve-like types, e.g. ``'a_bc'`` or ``'ab_c'``,
        or weak orders, e.g. ``'a~b>c'``. Default: all twelve-like types.
    d_type_fixed_share : dict, optional
        A dictionary. For each entry ``type: fixed_share``, this type will have at least this fixed share. The total
        must be lower or equal to 1.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileTwelve`.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_profile = RandProfileTwelveUniform()
        >>> profile = rand_profile()
        >>> print(profile)
        <a_bc: 0.3834415188257777, a_cb: 0.013932411923787802, ab_c: 0.040213280513127, ac_b: 0.10729597173420435, \
b_ac: 0.0039303209304278885, b_ca: 0.04313073699501224, ba_c: 0.05394987214431912, bc_a: 0.06929525330576336, \
c_ab: 0.07653567171024511, c_ba: 0.07188975971894951, ca_b: 0.10004796269941518, cb_a: 0.03633723949897072> \
(Condorcet winner: a)

    Using the optional parameters:

        >>> rand_profile = RandProfileTwelveUniform(
        ...     types=['a_bc', 'ac_b'], d_type_fixed_share={'b_ac': 0.8, 'a~b>c': 0.1},
        ...     voting_rule=PLURALITY)
        >>> profile = rand_profile()
        >>> print(profile)
        <a_bc: 0.052889491975290435, ac_b: 0.04711050802470954, b_ac: 0.8, a~b>c: 0.1> (Condorcet winner: b) (Plurality)

    For more examples, cf. :class:`RandSimplexUniform`.
    """

    def __init__(self, types=None, d_type_fixed_share=None, **kwargs):
        if types is None:
            types = TWELVE_TYPES
        super().__init__(cls=ProfileTwelve, keys=types, d_key_fixed_share=d_type_fixed_share, **kwargs)



from poisson_approval.constants.constants import *
from poisson_approval.random_factories.RandSimplexUniform import RandSimplexUniform
from poisson_approval.utils.Util import initialize_random_seeds, rand_simplex
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal
from poisson_approval.profiles.ProfileHistogram import ProfileHistogram


class RandProfileHistogramUniform(RandSimplexUniform):
    """A random factory of histogram profiles (:class:`ProfileHistogram`) following the uniform distribution.

    Parameters
    ----------
    n_bins : int
        The number of bins in each histogram. Cf. :class:`ProfileHistogram`.
    orders : list, optional
        These orders will have a variable share. They can be rankings, e.g. ``'abc'``, or weak orders, e.g.
        ``'a~b>c'``. Default: all rankings.
    d_order_fixed_share : dict, optional.
        A dictionary. For each entry ``order: fixed_share``, this order will have at least this fixed share. The total
        must be lower or equal to 1.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileHistogram`.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_profile = RandProfileHistogramUniform(n_bins=3)
        >>> profile = rand_profile()
        >>> print(profile)
        <abc: 0.4236547993389047 [0.43758721 0.2083069  0.35410589], \
acb: 0.12122838365799216 [0.891773   0.07188976 0.03633724], \
bac: 0.0039303209304278885 [0.38344152 0.40828352 0.20827496], \
bca: 0.05394987214431912 [0.52889492 0.03914964 0.43195544], \
cab: 0.1124259903007756 [0.07103606 0.85456058 0.07440336], \
cba: 0.2848106336275805 [0.0202184 0.0669109 0.9128707]> (Condorcet winner: a)

    Using the optional parameters:

        >>> rand_profile = RandProfileHistogramUniform(
        ...     n_bins=3, orders=['abc', 'a~b>c'], d_order_fixed_share={'b>a~c': 0.5},
        ...     voting_rule=PLURALITY)
        >>> profile = rand_profile()
        >>> print(profile)
        <abc: 0.416309922773969 [0.77815675 0.0918554  0.12998785], a~b>c: 0.083690077226031, b>a~c: 0.5> \
(Condorcet winner: b) (Plurality)
    """

    def __init__(self, n_bins, orders=None, d_order_fixed_share=None, **kwargs):
        if orders is None:
            orders = RANKINGS
        self.n_bins = n_bins
        self.kwargs_histogram = kwargs
        super().__init__(cls=ProfileOrdinal, keys=orders, d_key_fixed_share=d_order_fixed_share)

    def __call__(self):
        profile_ordinal = super().__call__()
        d_ranking_histogram = {}
        for ranking in sorted(profile_ordinal.support_in_rankings):
            d_ranking_histogram[ranking] = rand_simplex(d=self.n_bins)
        return ProfileHistogram(d_ranking_share=profile_ordinal.d_ranking_share,
                                d_ranking_histogram=d_ranking_histogram,
                                d_weak_order_share=profile_ordinal.d_weak_order_share,
                                **self.kwargs_histogram)



import math
import random
import itertools
import numpy as np
import sympy as sp
from fractions import Fraction
from decimal import Decimal
from poisson_approval.constants.constants import *
from poisson_approval.utils.DictPrintingInOrder import DictPrintingInOrder
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros


def initialize_random_seeds(n=0):
    """Initialize the random seeds.

    Parameters
    ----------
    n : int
        The desired random seed. Default: 0.
    """
    random.seed(n)
    np.random.seed(n)


def rand_simplex(d=6):
    """Draw a random point in the simplex.

    Parameters
    ----------
    d : int
        Number of coordinates. In other words, we consider the simplex of dimension `d - 1`.

    Returns
    -------
    numpy.ndarray
        A numpy array of length `d`, whose sum is 1.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> rand_simplex(d=6)  # doctest: +SKIP
        array([0.4236548 , 0.12122838, 0.00393032, 0.05394987, 0.11242599, 0.28481063])
    """
    x = np.sort(np.random.rand(d - 1))
    return np.concatenate((x, [1])) - np.concatenate(([0], x))


def rand_integers_fixed_sum(d, fixed_sum):
    """Generate integers with a given sum (uniformly).

    Parameters
    ----------
    d : int
        The desired number of integers. In other words, we consider a simplex of dimension `d - 1`.
    fixed_sum : int
        The fixed sum.

    Returns
    -------
    numpy.ndarray
        A numpy array of `d` integers, whose sum is `fixed_sum`, and drawn uniformly.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> rand_integers_fixed_sum(d=6, fixed_sum=100)
        array([ 2, 23, 34,  0, 22, 19])
    """
    n_separators = d - 1
    separators = np.concatenate((
        [-1],
        np.sort(np.random.choice(fixed_sum + n_separators, n_separators, replace=False)),
        [fixed_sum + n_separators]))
    return np.array([up - down - 1 for down, up in zip(separators[:-1], separators[1:])])


def rand_simplex_grid(d, denominator):
    """Draw a random point in the simplex, with rational coordinates of a given denominator

    Parameters
    ----------
    d : int
        Number of coordinates. In other words, we consider the simplex of dimension `d - 1`.
    denominator : int
        The coordinates will be fractions with this denominator.

    Returns
    -------
    numpy.ndarray
        A numpy array of length `d`, whose coordinates are fractions of the given denominator, and whose sum is 1.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> rand_simplex_grid(d=3, denominator=100)
        array([Fraction(13, 50), Fraction(13, 20), Fraction(9, 100)], dtype=object)
    """
    return np.array([
        my_division(int(n), denominator)
        for n in rand_integers_fixed_sum(d=d, fixed_sum=denominator)])


def probability(factory, n_samples, test, conditional_on=None):
    """Probability that a random `something` meets some given test.

    Parameters
    ----------
    factory : callable or tuple
        This can be:

        * Either a callable that takes no input and that outputs a (random) `something`,
        * Or a tuple of such factories (cf. examples below).
    n_samples : int
        Number of samples.
    test : callable or tuple
        This can be:

        * Either a function that take as input(s) the output(s) of the factory(ies) and that returns a Boolean.
        * Or a tuple of such functions (cf. examples below).
    conditional_on : callable
        A function that take as input(s) the output(s) of the factory(ies) and that returns a Boolean.
        Default: always True.

    Returns
    -------
    float or tuple
        This can be:

        * Either the probability that the output(s) generated by `factory` meet(s) `test`, conditional on the fact
          that it meets `conditional_on`, based on a Monte-Carlo estimation of `n_samples` trials.
        * Either a tuple giving this probability for each member of `test`, when `test` is a tuple itself.

    Examples
    --------
    In this basic example with one factory, we estimate the probability that a random float between 0 and 1 is greater
    than .5, conditionally on being greater than .25:

        >>> initialize_random_seeds()
        >>> def rand_number():
        ...     return random.random()
        >>> probability(factory=rand_number, n_samples=1000,
        ...             test=lambda x: x > .5, conditional_on=lambda x: x > .25)
        0.661

    In this example with a tuple of factories, we estimate the probability that a random 2*2 matrix and a random vector
    of size 2, both with integer coefficients between -10 included and 11 excluded, have a dot product that is null,
    conditionally on not being null themselves:

        >>> initialize_random_seeds()
        >>> def rand_matrix():
        ...     return np.random.randint(-10, 11, (2, 2))
        >>> def rand_vector():
        ...     return np.random.randint(-10, 11, 2)
        >>> def test_dot_zero(matrix, vector):
        ...     return np.all(np.dot(matrix, vector) == 0)
        >>> def test_non_trivial(matrix, vector):
        ...     return not np.all(matrix == 0) and not np.all(vector == 0)
        >>> probability(factory=(rand_matrix, rand_vector), n_samples=10000,
        ...             test=test_dot_zero, conditional_on=test_non_trivial)
        0.0003

    In the following example, we estimate the probability that a random float between 0 and 1 is greater than .5,
    and the probability that it is greater than .75, conditionally on being greater than .25:

        >>> initialize_random_seeds()
        >>> def rand_number():
        ...     return random.random()
        >>> probability(factory=rand_number, n_samples=1000,
        ...             test=(lambda x: x > .5, lambda x: x > .75), conditional_on=lambda x: x > .25)
        (0.661, 0.332)

    When using a tuple of tests, the same samples are used to estimate each probability.
    """
    if not isinstance(factory, tuple):
        factory = (factory,)
    is_test_tuple = isinstance(test, tuple)
    if not is_test_tuple:
        test = (test,)
    l_test_success = [0 for the_test in test]
    i_samples = 0
    while i_samples < n_samples:
        somethings = [f() for f in factory]
        if conditional_on is None or conditional_on(*somethings):
            i_samples += 1
            for i_test, the_test in enumerate(test):
                if the_test(*somethings):
                    l_test_success[i_test] += 1
    l_test_rate = [successes / n_samples for successes in l_test_success]
    if is_test_tuple:
        return tuple(l_test_rate)
    else:
        return l_test_rate[0]


def image_distribution(factory, n_samples, f, conditional_on=None):
    """Distribution of ``f(something)`` for a random `something`.

    Parameters
    ----------
    factory : callable or tuple
        This can be:

        * Either a callable that takes no input and that outputs a (random) `something`,
        * Or a tuple of such factories (cf. examples below).
    n_samples : int
        Number of samples.
    f : callable
        A function that take as input(s) the output(s) of the factory(ies).
    conditional_on : callable
        A function that take as input(s) the output(s) of the factory(ies) and that returns a Boolean.
        Default: always True.

    Returns
    -------
    DictPrintingInOrder
        Keys: the obtained outputs for `f`. Values: the probability that `f(something)` has this output when `something`
        is generated by `factory`, conditional on the fact that it meets `conditional_on`,
        based on a Monte-Carlo estimation of `n_samples` trials.

    Examples
    --------
    In this basic example with one factory, we compute the distribution of `n` modulo 10, when `n` is drawn uniformly
    at random between 0 included and 100 excluded, conditionally on the fact that `n` is even:

        >>> initialize_random_seeds()
        >>> def rand_integer():
        ...     return np.random.randint(0, 100)
        >>> def modulo_10(n):
        ...     return n % 10
        >>> def is_even(n):
        ...     return n % 2 == 0
        >>> image_distribution(factory=rand_integer, n_samples=100, f=modulo_10, conditional_on=is_even)
        {0: 0.21, 2: 0.21, 4: 0.2, 6: 0.18, 8: 0.2}

    In this example with a tuple of factories, we compute the distribution of `a mod b`, when `a` is drawn uniformly
    at random between 0 included and 100 excluded, and `b` is drawn uniformly at random between 1 included and 11
    excluded:

        >>> initialize_random_seeds()
        >>> def rand_integer():
        ...     return np.random.randint(0, 100)
        >>> def rand_divider():
        ...     return np.random.randint(1, 11)
        >>> def modulo(a, b):
        ...     return a % b
        >>> image_distribution(factory=(rand_integer, rand_divider),
        ...                    n_samples=100, f=modulo)
        {0: 0.31, 1: 0.16, 2: 0.18, 3: 0.12, 4: 0.07, 5: 0.04, 6: 0.02, 7: 0.08, 9: 0.02}
    """
    if not isinstance(factory, tuple):
        factory = (factory,)
    d_result_occurrences = dict()
    i_samples = 0
    while i_samples < n_samples:
        somethings = [g() for g in factory]
        if conditional_on is None or conditional_on(*somethings):
            i_samples += 1
            result = f(*somethings)
            d_result_occurrences[result] = d_result_occurrences.get(result, 0) + 1
    return DictPrintingInOrder({result: occurrences / n_samples
                                for result, occurrences in d_result_occurrences.items()})


def _false_for_fraction(f):
    """Decorator to return False when the input is a Fraction (cf. usages below)."""
    def _f(x):
        if isinstance(x, Fraction):
            return False
        return f(x)
    _f.__doc__ = f.__doc__
    return _f


@_false_for_fraction
def isnan(x):
    """Is nan.

    Parameters
    ----------
    x : Number

    Returns
    -------
    bool
        True if `x` is nan.

    Notes
    -----
    This extends the usual numpy function ``isnan`` to fractions and sympy expressions.

    Examples
    --------
        >>> values = [sp.sqrt(3) - sp.sqrt(2), sp.nan,
        ...           sp.oo, - sp.oo,
        ...           sp.Rational(3, 5), Fraction(3, 5),
        ...           1, 0.42, np.inf, -np.inf, np.nan]
        >>> print([x for x in values if isnan(x)])
        [nan, nan]
    """
    if isinstance(x, sp.Expr):
        return x == sp.nan
    else:
        return np.isnan(x)


@_false_for_fraction
def isposinf(x):
    """Is positive infinity.

    Parameters
    ----------
    x : Number

    Returns
    -------
    bool
        True if `x` is positive infinity.

    Notes
    -----
    This extends the usual numpy function ``isposinf`` to fractions and sympy expressions.

    Examples
    --------
        >>> values = [sp.sqrt(3) - sp.sqrt(2), sp.nan,
        ...           sp.oo, - sp.oo,
        ...           sp.Rational(3, 5), Fraction(3, 5),
        ...           1, 0.42, np.inf, -np.inf, np.nan]
        >>> print([x for x in values if isposinf(x)])
        [oo, inf]
    """
    if isinstance(x, sp.Expr):
        return x == sp.oo
    else:
        return np.isposinf(x)


@_false_for_fraction
def isneginf(x):
    """Is negative infinity.

    Parameters
    ----------
    x : Number

    Returns
    -------
    bool
        True if `x` is negative infinity.

    Notes
    -----
    This extends the usual numpy function ``isneginf`` to fractions and sympy expressions.

    Examples
    --------
        >>> values = [sp.sqrt(3) - sp.sqrt(2), sp.nan,
        ...           sp.oo, - sp.oo,
        ...           sp.Rational(3, 5), Fraction(3, 5),
        ...           1, 0.42, np.inf, -np.inf, np.nan]
        >>> print([x for x in values if isneginf(x)])
        [-oo, -inf]
    """
    if isinstance(x, sp.Expr):
        return x == - sp.oo
    else:
        return np.isneginf(x)


def give_figure(n, singular, plural=None):
    """Combine a number with a unit, whose word can be singular or plural.

    Parameters
    ----------
    n : int
        The number.
    singular : str
        The singular word.
    plural : str
        The plural word. Default: the singular with a final 's'.

    Returns
    -------
    str
        The number and the word.

    Examples
    --------
        >>> give_figure(1, 'apple')
        '1 apple'
        >>> give_figure(2, 'apple')
        '2 apples'
        >>> give_figure(1, 'equilibrium', 'equilibria')
        '1 equilibrium'
        >>> give_figure(2, 'equilibrium', 'equilibria')
        '2 equilibria'
    """
    if n <= 1:
        return str(n) + ' ' + singular
    else:
        if plural is None:
            return str(n) + ' ' + singular + 's'
        else:
            return str(n) + ' ' + plural


def to_callable(o):
    """Convert to a callable.

    Parameters
    ----------
    o : object

    Returns
    -------
    callable
        The conversion of `o` to a callable.

    Examples
    --------
    If `o` is callable, then return `o`:

        >>> def square(x):
        ...     return x**2
        >>> my_function = to_callable(square)
        >>> my_function(4)
        16

    If `o` is not callable, then return a function ``*args, **kwargs -> o``:

        >>> x = 42
        >>> my_function = to_callable(x)
        >>> my_function('some_argument', keyword='some_value')
        42
    """
    if callable(o):
        return o
    else:
        def my_function(*args, **kwargs):
            return o
        return my_function


def product_dict(d_key_possible_values):
    """Iterable: product of dictionaries.

    Source: https://stackoverflow.com/questions/5228158/cartesian-product-of-a-dictionary-of-lists.

    Parameters
    ----------
    d_key_possible_values
        To each key, associate a list of possible values (cf. example below).

    Yields
    ------
    dict
        A dictionary that, to each key, associates one of its possible values. All elements of the Cartesian product
        are returned this way.

    Examples
    --------
        >>> d_key_possible_values = {'foo': [0, 1], 'bar': ['a', 'b', 'c']}
        >>> for d_key_value in product_dict(d_key_possible_values):
        ...     print(d_key_value)
        {'foo': 0, 'bar': 'a'}
        {'foo': 0, 'bar': 'b'}
        {'foo': 0, 'bar': 'c'}
        {'foo': 1, 'bar': 'a'}
        {'foo': 1, 'bar': 'b'}
        {'foo': 1, 'bar': 'c'}
    """
    keys = d_key_possible_values.keys()
    vals = d_key_possible_values.values()
    for instance in itertools.product(*vals):
        yield dict(zip(keys, instance))


def candidates_to_d_candidate_probability(candidates):
    """Convert a set of candidates to a dictionary of probabilities (random tie-break).

    Parameters
    ----------
    candidates : set
        A subset of ``{'a', 'b', 'c'}``. Typically: a set of winners.

    Returns
    -------
    DictPrintingInOrderIgnoringZeros
        Key: ``'a'``, ``'b'`` or ``'c'``. Value: the probability of this candidate winning with a uniformly random
        tie-break.

    Examples
    --------
        >>> winners = {'a', 'b'}
        >>> candidates_to_d_candidate_probability(winners)
        {'a': Fraction(1, 2), 'b': Fraction(1, 2)}
    """
    n_candidates = len(candidates)
    return DictPrintingInOrderIgnoringZeros({
        candidate: Fraction(1, n_candidates) if candidate in candidates else 0 for candidate in CANDIDATES})


def candidates_to_probabilities(candidates):
    """Convert a set of candidates to an array of probabilities (random tie-break).

    Parameters
    ----------
    candidates : set
        A subset of ``{'a', 'b', 'c'}``. Typically: a set of winners.

    Returns
    -------
    ndarray
        Array of size 3. For example, the first coefficient is the probability that candidate `a` win by a random
        tie-break.

    Examples
    --------
        >>> winners = {'a', 'b'}
        >>> candidates_to_probabilities(winners)
        array([Fraction(1, 2), Fraction(1, 2), 0], dtype=object)
    """
    n_candidates = len(candidates)
    return np.array([Fraction(1, n_candidates) if candidate in candidates else 0 for candidate in CANDIDATES])


def array_to_d_candidate_value(values):
    """Convert an array to a dictionary of candidates and values

    Parameters
    ----------
    values : ndarray
        An array of size 3.

    Returns
    -------
    DictPrintingInOrderIgnoringZeros
        The corresponding dictionary.

    Examples
    --------
        >>> values = [42, 51, 69]
        >>> array_to_d_candidate_value(values)
        {'a': 42, 'b': 51, 'c': 69}
    """
    return DictPrintingInOrderIgnoringZeros({candidate: value for candidate, value in zip(CANDIDATES, values)})


def d_candidate_value_to_array(d_candidate_value):
    """Convert a dictionary of candidates and values to an array

    Parameters
    ----------
    d_candidate_value : dict
        Key: ``'a'``, ``'b'`` or ``'c'``.

    Returns
    -------
    ndarray
        The corresponding array.

    Examples
    --------
        >>> d_candidate_value = {'a': 42, 'b': 51, 'c': 69}
        >>> d_candidate_value_to_array(d_candidate_value)
        array([42, 51, 69])
    """
    return np.array([d_candidate_value[candidate] for candidate in CANDIDATES])


def one_over_t(t):
    """Function `1 / t`.

    When used as an update ratio (cf. :meth:`ProfileCardinal.fictitious_play`), this amounts to computing the arithmetic
    mean.

    Parameters
    ----------
    t : Number

    Returns
    -------
    Number

    Examples
    --------
        >>> one_over_t(2)
        Fraction(1, 2)
    """
    return my_division(1, t)


def one_over_sqrt_t(t):
    """Function `1 / sqrt(t)`.

    This function is provided as an example of update ratio for :meth:`ProfileCardinal.fictitious_play`.

    Parameters
    ----------
    t : Number

    Returns
    -------
    Number

    Examples
    --------
        >>> one_over_sqrt_t(2)
        0.7071067811865475
    """
    return my_division(1, math.sqrt(t))


def one_over_log_t_plus_one(t):
    """Function `1 / log(t + 1)`.

    This function is provided as an example of update ratio for :meth:`ProfileCardinal.fictitious_play`. The constant
    1 in the denominator is the smallest integer such that `f(t = 2) < 1`.

    Parameters
    ----------
    t : Number

    Returns
    -------
    Number

    Examples
    --------
        >>> one_over_log_t_plus_one(2)
        0.9102392266268373
    """
    return my_division(1, math.log(t + 1))


def one_over_log_log_t_plus_fourteen(t):
    """Function `1 / log(log(t + 14))`.

    This function is provided as an example of update ratio for :meth:`ProfileCardinal.fictitious_play`. The constant
    14 in the denominator is the smallest integer such that `f(t = 1) < 1`.

    Parameters
    ----------
    t : Number

    Returns
    -------
    Number

    Examples
    --------
        >>> one_over_log_log_t_plus_fourteen(2)
        0.9806022744169713
    """
    return my_division(1, math.log(math.log(t + 14)))


def my_division(x, y):
    """
    Division of two numbers, trying to be exact if it is reasonable.

    Parameters
    ----------
    x : Number
    y : Number

    Returns
    -------
    Number
        The division of `x` by `y`.

    Examples
    --------
    Typical usages:

        >>> my_division(6, 2)
        3
        >>> my_division(5, 2)
        Fraction(5, 2)

    If `x` or `y` is a float, then the result is a float:

        >>> my_division(Fraction(5, 2), 0.1)
        25.0
        >>> my_division(0.1, Fraction(5, 2))
        0.04

    If `x` and `y` are integers, decimals, fractions or sympy expressions, then the result is symbolic:

        >>> my_division(2, Fraction(5, 2))
        Fraction(4, 5)
        >>> my_division(Decimal('0.1'), Fraction(5, 2))
        Fraction(1, 25)
        >>> my_division(sp.sqrt(3), 2)
        sqrt(3)/2

    Possible errors:

        >>> my_division(1, 0)
        Traceback (most recent call last):
        ZeroDivisionError: division by zero
        >>> my_division(1, 'foo')
        Traceback (most recent call last):
        TypeError: unsupported operand type(s) for /: 'Fraction' and 'str'

    """
    if y == 0:
        raise ZeroDivisionError('division by zero')
    if isinstance(x, float) or isinstance(y, float):
        return x / y
    if not isinstance(x, sp.Rational):
        try:
            x = Fraction(x)
        except (TypeError, ValueError):
            pass
    if not isinstance(y, sp.Rational):
        try:
            y = Fraction(y)
        except (TypeError, ValueError):
            pass
    result = x / y
    if isinstance(result, Fraction) and result.denominator == 1:
        return result.numerator
    else:
        return result


def iterator_integers_fixed_sum(d, fixed_sum):
    """Iterate over vectors of integers with a fixed sum.

    Parameters
    ----------
    d : int
        The desired number of integers. In other words, we consider a simplex of dimension `d - 1`.
    fixed_sum : int
        The fixed sum.

    Yields
    ------
    tuple
        Each tuple of `d` integers, whose sum is `fixed_sum`.

    Examples
    --------
        >>> for t in iterator_integers_fixed_sum(d=3, fixed_sum=2):
        ...     print(t)
        (2, 0, 0)
        (1, 1, 0)
        (1, 0, 1)
        (0, 2, 0)
        (0, 1, 1)
        (0, 0, 2)
    """
    if d == 1:
        yield fixed_sum,
    else:
        for i in range(fixed_sum, -1, -1):
            for r in iterator_integers_fixed_sum(d - 1, fixed_sum - i):
                yield (i, *r)


def iterate_simplex_grid(d, denominator):
    """Iterate over the points in the simplex, with rational coordinates of a given denominator

    Parameters
    ----------
    d : int
        Number of coordinates. In other words, we consider the simplex of dimension `d - 1`.
    denominator : int or iterable
        The coordinates will be fractions with this denominator. If an iterable is given, we consider each
        denominator given by the iterable.

    Returns
    -------
    tuple
        Each tuple of length `d`, whose coordinates are fractions of the given denominator, and whose sum is 1.

    Examples
    --------
        >>> for t in iterate_simplex_grid(d=3, denominator=range(1, 3)):
        ...     print(t)
        (1, 0, 0)
        (0, 1, 0)
        (0, 0, 1)
        (1, 0, 0)
        (Fraction(1, 2), Fraction(1, 2), 0)
        (Fraction(1, 2), 0, Fraction(1, 2))
        (0, 1, 0)
        (0, Fraction(1, 2), Fraction(1, 2))
        (0, 0, 1)
    """
    if isinstance(denominator, int):
        denominators = [denominator]
    else:
        denominators = denominator
    for current_denominator in denominators:
        for t in iterator_integers_fixed_sum(d, fixed_sum=current_denominator):
            yield tuple(my_division(x, current_denominator) for x in t)


def my_range(start, end, step):
    """Iterable `range` adapted for fractions.

    Parameters
    ----------
    start : Number
        Start value (included).
    end : Number
        End value (excluded).
    step : Number
        Increment of the counter.

    Examples
    --------
        >>> for x in my_range(0, 1, Fraction(1, 3)):
        ...     print(x)
        0
        1/3
        2/3

        >>> for x in my_range(1, 0, - Fraction(1, 3)):
        ...     print(x)
        1
        2/3
        1/3
    """
    val = start
    if step > 0:
        while val < end:
            yield val
            val += step
    else:
        while val > end:
            yield val
            val += step


def my_sign(x):
    """Sign.

    Parameters
    ----------
    x : Number

    Returns
    -------
    int
        Sign of x.

    Examples
    --------
        >>> my_sign(1.5)
        1
        >>> my_sign(0)
        0
        >>> my_sign(-4.2)
        -1
    """
    if x > 0:
        return 1
    elif x < 0:
        return -1
    else:
        return 0



from poisson_approval.utils.Util import isneginf
from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event


class EventTrio2t(Event):
    """A 3-candidate almost-tie of the type (X - 1, X - 1, X).

    Notes
    -----
    We consider the trios of type 1, i.e. situations where ``S_x + 1 = S_y + 1 = S_z``.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> from poisson_approval import TauVector
        >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(6, 10), 'c': Fraction(3, 10)})
        >>> EventTrio2t(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 1.18339 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        ce = self.ce
        trio = self.tau.trio
        self._phi_x = trio.phi[self._label_x]
        self._phi_y = trio.phi[self._label_y]
        self._phi_z = trio.phi[self._label_z]
        self._phi_xy = trio.phi[self._label_xy]
        self._phi_xz = trio.phi[self._label_xz]
        self._phi_yz = trio.phi[self._label_yz]
        if tau_z == 0 and tau_xy == 0:
            # Cross diagram 1
            self.asymptotic = (Asymptotic.poisson_one_more(tau_yz, tau_x, symbolic=self.symbolic)
                               * Asymptotic.poisson_one_more(tau_xz, tau_y, symbolic=self.symbolic))
        elif (tau_x == 0 and tau_yz == 0) or (tau_y == 0 and tau_xz == 0):
            # Cross diagram 2
            self.asymptotic = (Asymptotic.poisson_eq(tau_x, tau_yz, symbolic=self.symbolic)
                               * Asymptotic.poisson_eq(tau_y, tau_xz, symbolic=self.symbolic)
                               * Asymptotic.poisson_one_more(tau_z, tau_xy, symbolic=self.symbolic))
        elif (tau_z == 0 and tau_xz == 0) or (tau_z == 0 and tau_yz == 0):
            # Flower diagram 1
            self.asymptotic = Asymptotic(mu=-ce.inf, nu=-ce.inf, xi=-ce.inf, symbolic=self.symbolic)
        elif (tau_x == 0 and tau_xz == 0) or (tau_y == 0 and tau_yz == 0):
            # Flower diagram 2
            self.asymptotic = (Asymptotic.poisson_eq(tau_x, tau_yz, symbolic=self.symbolic)
                               * Asymptotic.poisson_eq(tau_y, tau_xz, symbolic=self.symbolic)
                               * Asymptotic.poisson_one_more(tau_z, tau_xy, symbolic=self.symbolic))
        elif (tau_x == 0 and tau_xy == 0) or (tau_y == 0 and tau_xy == 0):
            # Flower diagram 3
            self.asymptotic = (
                (Asymptotic.poisson_value(tau_z, 0, symbolic=self.symbolic)
                 * Asymptotic.poisson_one_more(tau_yz, tau_x, symbolic=self.symbolic)
                 * Asymptotic.poisson_one_more(tau_xz, tau_y, symbolic=self.symbolic))
                + (Asymptotic.poisson_one_more(tau_z, 1, symbolic=self.symbolic)
                   * Asymptotic.poisson_eq(tau_x, tau_yz, symbolic=self.symbolic)
                   * Asymptotic.poisson_eq(tau_y, tau_xz, symbolic=self.symbolic))
            )
        else:
            psi_xy = trio.psi[self._label_xy]
            self.asymptotic = trio.asymptotic * psi_xy
        if isneginf(self.asymptotic.mu):
            self._phi_x = ce.nan
            self._phi_y = ce.nan
            self._phi_z = ce.nan
            self._phi_xy = ce.nan
            self._phi_xz = ce.nan
            self._phi_yz = ce.nan



class DictPrintingInOrderIgnoringNone(dict):
    """A dictionary that prints in the order of the keys, ignoring value None (or iterables containing only None).

    Examples
    --------
        >>> d = DictPrintingInOrderIgnoringNone({'b': 'x', 'a': 'y', 'c': 'z', 'd': 0, 'e': None, 'f': [None, None]})
        >>> print(d)
        {a: y, b: x, c: z, d: 0}
        >>> print(repr(d))
        {'a': 'y', 'b': 'x', 'c': 'z', 'd': 0}
    """

    def __repr__(self):
        return "{" + ", ".join([
            "%r: %r" % (key, self[key]) for key in sorted(self)
            if interesting_value(self[key])
        ]) + "}"

    def __str__(self):
        return "{" + ", ".join([
            "%s: %s" % (key, self[key]) for key in sorted(self)
            if interesting_value(self[key])
        ]) + "}"

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')


def interesting_value(value):
    if value is None:
        return False
    try:
        if all(element is None for element in value):
            return False
    except TypeError:
        return True
    return True



from poisson_approval.profiles.ProfileCardinal import ProfileCardinal
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold


# noinspection PyAbstractClass
class ProfileCardinalContinuous(ProfileCardinal):
    """A cardinal profile of preference with no atom (abstract class).

    Parameters
    ----------
    Cf. :class:`ProfileCardinal`.
    """

    is_continuous = True

    def have_ranking_with_utility_u(self, ranking, u):
        """Share of voters who have a given ranking and a given utility for their middle candidate.

        Since it is a continuous profile, this method always returns 0.
        """
        return 0

    def best_responses_to_strategy(self, d_ranking_best_response):
        """Convert best responses to a :class:`StrategyThreshold`.

        Parameters
        ----------
        d_ranking_best_response : dict
            Key: ranking. Value: :class:`BestResponse`.

        Returns
        -------
        StrategyThreshold
            The conversion of the best responses into a strategy. Only the rankings present in this profile are
            mentioned in the strategy.
        """
        # The only difference with parent class is that ratio_optimistic is not specified because we don't care.
        return StrategyThreshold({
            ranking: best_response.threshold_utility
            for ranking, best_response in d_ranking_best_response.items()
            if self.d_ranking_share[ranking] > 0
        }, profile=self, voting_rule=self.voting_rule)



from poisson_approval.constants.constants import *
from poisson_approval.random_factories.RandSimplexUniform import RandSimplexUniform
from poisson_approval.utils.Util import initialize_random_seeds
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal


class RandProfileOrdinalUniform(RandSimplexUniform):
    """A random factory of ordinal profiles (:class:`ProfileOrdinal`) following the uniform distribution.

    Parameters
    ----------
    orders : list, optional
        These orders will have a variable share. They can be rankings, e.g. ``'abc'``, or weak orders, e.g.
        ``'a~b>c'``. Default: all rankings.
    d_order_fixed_share : dict, optional.
        A dictionary. For each entry ``order: fixed_share``, this order will have at least this fixed share. The total
        must be lower or equal to 1.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileOrdinal`.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_profile = RandProfileOrdinalUniform()
        >>> profile = rand_profile()
        >>> print(profile)
        <abc: 0.4236547993389047, acb: 0.12122838365799216, bac: 0.0039303209304278885, bca: 0.05394987214431912, \
cab: 0.1124259903007756, cba: 0.2848106336275805> (Condorcet winner: a)

    Using the optional parameters:

        >>> rand_profile = RandProfileOrdinalUniform(
        ...     orders=['abc', 'acb'], d_order_fixed_share={'acb': 0.8, 'a~b>c': 0.1},
        ...     voting_rule=PLURALITY)
        >>> profile = rand_profile()
        >>> print(profile)
        <abc: 0.0645894113066656, acb: 0.8354105886933344, a~b>c: 0.1> (Condorcet winner: a) (Plurality)

    For more examples, cf. :class:`RandSimplexUniform`.
    """

    def __init__(self, orders=None, d_order_fixed_share=None, **kwargs):
        if orders is None:
            orders = RANKINGS
        super().__init__(cls=ProfileOrdinal, keys=orders, d_key_fixed_share=d_order_fixed_share, **kwargs)



from fractions import Fraction
from poisson_approval import XyyToProfile, ProfileNoisyDiscrete


def test_without_fixed_voters():
    """
        >>> xyy_to_profile = XyyToProfile(
        ...     ProfileNoisyDiscrete, left_ranking='abc', right_ranking='bac', noise=0.01)
        >>> profile = xyy_to_profile(x=Fraction(3, 10), y1=0.42, y2=0.51)
        >>> print(profile)
        <abc 0.42 ± 0.01: 7/10, bac 0.51 ± 0.01: 3/10> (Condorcet winner: a)
    """
    pass



import random
from poisson_approval.constants.constants import *
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.strategies.StrategyOrdinal import StrategyOrdinal
from poisson_approval.utils.Util import initialize_random_seeds
from poisson_approval.utils.UtilBallots import ballot_low_u, ballot_high_u


class RandStrategyOrdinalUniform:
    """A random factory of ordinal strategies (:class:`StrategyOrdinal`) following the uniform distribution.

    Parameters
    ----------
    profile : Profile, optional
    voting_rule : str, optional
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``. Default: the same
        voting rule as `profile` if a profile is specified, ``APPROVAL`` otherwise.
    d_ranking_fixed_strategy : dict
        Key: ranking. Value: fixed strategy. Cf. examples below.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_strategy = RandStrategyOrdinalUniform()
        >>> strategy = rand_strategy()
        >>> print(strategy)
        <abc: ab, acb: ac, bac: b, bca: bc, cab: ac, cba: bc>

    Specify a profile:

        >>> from poisson_approval import ProfileOrdinal
        >>> profile = ProfileOrdinal({'abc': 0.75, 'bac': 0.25})
        >>> rand_strategy = RandStrategyOrdinalUniform(profile=profile)
        >>> strategy = rand_strategy()
        >>> print(strategy)
        <abc: ab, bac: ab> ==> a, b

    Specify some fixed strategies:

        >>> rand_strategy = RandStrategyOrdinalUniform(
        ...     d_ranking_fixed_strategy={'abc': 'a', 'acb': 'a', 'bac': 'b', 'bca': 'b'})
        >>> print(rand_strategy())
        <abc: a, acb: a, bac: b, bca: b, cab: ac, cba: c>
        >>> print(rand_strategy())
        <abc: a, acb: a, bac: b, bca: b, cab: ac, cba: bc>
        >>> print(rand_strategy())
        <abc: a, acb: a, bac: b, bca: b, cab: ac, cba: c>
        >>> print(rand_strategy())
        <abc: a, acb: a, bac: b, bca: b, cab: ac, cba: bc>
        >>> print(rand_strategy())
        <abc: a, acb: a, bac: b, bca: b, cab: c, cba: c>
    """

    def __init__(self, profile=None, voting_rule=None, d_ranking_fixed_strategy=None):
        # Default parameters
        if d_ranking_fixed_strategy is None:
            d_ranking_fixed_strategy = dict()
        # Parameters
        self.profile = profile
        # noinspection PyProtectedMember
        self.voting_rule = Strategy._get_voting_rule_(profile, voting_rule)
        self.d_ranking_fixed_strategy = d_ranking_fixed_strategy

        # Computed variables

        def possible_ballots(ranking):
            try:
                return [self.d_ranking_fixed_strategy[ranking]]
            except KeyError:
                pass
            if self.profile is not None and self.profile.d_ranking_share[ranking] == 0:
                return ['']
            return [ballot_low_u(ranking, self.voting_rule), ballot_high_u(ranking, self.voting_rule)]

        self.d_ranking_possible_ballots = {ranking: possible_ballots(ranking) for ranking in RANKINGS}

    def __call__(self):
        return StrategyOrdinal(
            d_ranking_ballot={ranking: random.choice(possible_ballots)
                              for ranking, possible_ballots in self.d_ranking_possible_ballots.items()},
            profile=self.profile,
            voting_rule=self.voting_rule
        )



import random
import numpy as np
from fractions import Fraction
from poisson_approval.constants.constants import *
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.containers.AnalyzedStrategies import AnalyzedStrategies
from poisson_approval.containers.Winners import Winners
from poisson_approval.iterables.IterableStrategyOrdinal import IterableStrategyOrdinal
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.utils.computation_engine import computation_engine
from poisson_approval.utils.SetPrintingInOrder import SetPrintingInOrder
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.utils.Util import my_division
from poisson_approval.utils.UtilPreferences import is_lover
from poisson_approval.utils.UtilBallots import sort_ballot, ballot_high_u, ballot_low_u
from poisson_approval.utils.UtilCache import cached_property, DeleteCacheMixin, property_deleting_cache


# noinspection PyUnresolvedReferences
class Profile(DeleteCacheMixin):
    """A profile of preference (abstract class).

    Parameters
    ----------
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.
    symbolic : bool
        Whether the computations are symbolic or numeric.
    """

    def __init__(self, voting_rule=APPROVAL, symbolic=False):
        self.voting_rule = voting_rule
        self.symbolic = symbolic
        self.ce = computation_engine(symbolic)

    voting_rule = property_deleting_cache('_voting_rule')

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    @cached_property
    def d_ranking_share(self):
        """dict : Shares of rankings.
            E.g. ``'abc': 0.3`` means that a ratio 0.3 of the voters have ranking ``'abc'``.
        """
        raise NotImplementedError

    @cached_property
    def d_weak_order_share(self):
        """dict : Shares of weak orders.
            E.g. ``'a~b>c': 0.3`` means that a ratio 0.3 of the voters have weak order ``'a~b>c'``.
        """
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError

    @cached_property
    def standardized_version(self):
        """Profile : Standardized version of the profile (makes it unique, up to permutations of the candidates)."""
        raise NotImplementedError

    @cached_property
    def is_standardized(self):
        """bool : Whether the profile is standardized. Cf. :meth:`standardized_version`."""
        return self == self.standardized_version

    # Condorcet stuff

    @cached_property
    def weighted_maj_graph(self):
        """np.ndarray : Weighted majority graph."""
        ab = (self.abc + self.acb + self.cab - self.bac - self.bca - self.cba
              + self.d_weak_order_share['a>b~c'] - self.d_weak_order_share['b~c>a']
              + self.d_weak_order_share['a~c>b'] - self.d_weak_order_share['b>a~c'])
        ac = (self.abc + self.acb + self.bac - self.bca - self.cab - self.cba
              + self.d_weak_order_share['a>b~c'] - self.d_weak_order_share['b~c>a']
              + self.d_weak_order_share['a~b>c'] - self.d_weak_order_share['c>a~b'])
        bc = (self.abc + self.bac + self.bca - self.acb - self.cab - self.cba
              + self.d_weak_order_share['b>a~c'] - self.d_weak_order_share['a~c>b']
              + self.d_weak_order_share['a~b>c'] - self.d_weak_order_share['c>a~b'])
        return np.array([[0, ab, ac], [-ab, 0, bc], [-ac, -bc, 0]])

    @cached_property
    def condorcet_winners(self):
        """Winners : Condorcet winner(s)."""
        m = self.weighted_maj_graph
        min_score = {'a': min(m[0, 1], m[0, 2]), 'b': min(m[1, 0], m[1, 2]), 'c': min(m[2, 0], m[2, 1])}
        return Winners({candidate for candidate in ['a', 'b', 'c'] if min_score[candidate] > - 10**(-8)})

    @cached_property
    def is_profile_condorcet(self):
        """float : Whether the profile is Condorcet. 1. means there is a strict Condorcet winner, 0.5 means there are
        one or more weak Condorcet winner(s), 0. means there is no Condorcet winner.
        """
        m = self.weighted_maj_graph
        min_score = [min(m[0, 1], m[0, 2]), min(m[1, 0], m[1, 2]), min(m[2, 0], m[2, 1])]
        maxi_min = max(min_score)
        if self.ce.look_equal(maxi_min, 0, abs_tol=1E-8):
            return .5
        elif maxi_min > 0:
            return 1.
        else:
            return 0.

    @cached_property
    def has_majority_favorite(self):
        """bool : Whether there is a `majority favorite` (a candidate ranked first by strictly more than half of the
        voters).
        """
        return (self.abc + self.acb + self.d_weak_order_share['a>b~c'] > Fraction(1, 2)
                or self.bac + self.bca + self.d_weak_order_share['b>a~c'] > Fraction(1, 2)
                or self.cab + self.cba + self.d_weak_order_share['c>a~b'] > Fraction(1, 2))

    @cached_property
    def has_majority_ranking(self):
        """bool : Whether there is a majority ranking (a ranking shared by strictly more than half of the voters)."""
        return max(self.d_ranking_share.values()) > Fraction(1, 2)

    # Single-peakedness
    @cached_property
    def is_single_peaked(self):
        """bool : Whether the profile is single-peaked."""
        return ((self.abc == 0 and self.bac == 0 and self.d_weak_order_share['a~b>c'] == 0
                 and self.d_weak_order_share['a>b~c'] == 0 and self.d_weak_order_share['b>a~c'] == 0)
                or (self.acb == 0 and self.cab == 0 and self.d_weak_order_share['a~c>b'] == 0
                    and self.d_weak_order_share['a>b~c'] == 0 and self.d_weak_order_share['c>a~b'] == 0)
                or (self.bca == 0 and self.cba == 0 and self.d_weak_order_share['b~c>a'] == 0
                    and self.d_weak_order_share['c>a~b'] == 0 and self.d_weak_order_share['b>a~c'] == 0))

    # Has full support
    @cached_property
    def support_in_rankings(self):
        """:class:`SetPrintingInOrder` of str : Support of the profile (in terms of rankings)."""
        return SetPrintingInOrder({key for key, val in self.d_ranking_share.items() if val > 0})

    @cached_property
    def is_generic_in_rankings(self):
        """bool : Whether the profile is generic in rankings (contains all rankings)."""
        return 0 not in self.d_ranking_share.values()

    @cached_property
    def contains_rankings(self):
        """bool : Whether the profile contains some rankings."""
        return len(self.support_in_rankings) > 0

    @cached_property
    def support_in_weak_orders(self):
        """:class:`SetPrintingInOrder` of str : Support of the profile (in terms of weak orders)."""
        return SetPrintingInOrder({key for key, val in self.d_weak_order_share.items() if val > 0})

    @cached_property
    def contains_weak_orders(self):
        """bool : Whether the profile contains some weak orders."""
        return len(self.support_in_weak_orders) > 0

    # Tau and strategy-related stuff

    def tau(self, strategy):
        """Tau-vector associated to this profile and a given strategy.

        Parameters
        ----------
        strategy : Strategy
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.
        """
        raise NotImplementedError

    # noinspection NonAsciiCharacters
    def τ(self, strategy):
        """Tau-vector (alternate notation).

        Parameters
        ----------
        strategy : Strategy
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.
        """
        return self.tau(strategy)

    @cached_property
    def d_ballot_share_weak_voters_fanatic(self):
        """dict : Ballot shares due to the weak orders if they vote fanatically

        Voters of the type ``'a>b~c'``:

        * In Approval or Plurality, they vote for `a`.
        * In Anti-plurality, half of them vote for `ab` (i.e. against `c`) and half of them vote for `ac` (i.e.
          against `b`).

        Voters of the type ``'a~b>c'``:

        * In Approval or Plurality, half of them vote for `a` and half of them vote for `b`.
        * In Anti-plurality, they vote for `ab` (i.e. against `c`).
        """
        d = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for weak_order, share in self.d_weak_order_share.items():
            if is_lover(weak_order):
                if self.voting_rule in {APPROVAL, PLURALITY}:
                    d[weak_order[0]] += share
                elif self.voting_rule == ANTI_PLURALITY:
                    d[sort_ballot(weak_order[0] + weak_order[2])] += my_division(share, 2)
                    d[sort_ballot(weak_order[0] + weak_order[4])] += my_division(share, 2)
                else:
                    raise NotImplementedError
            else:  # is_hater(weak_order)
                if self.voting_rule in {APPROVAL, PLURALITY}:
                    d[weak_order[0]] += my_division(share, 2)
                    d[weak_order[2]] += my_division(share, 2)
                elif self.voting_rule == ANTI_PLURALITY:
                    d[sort_ballot(weak_order[0] + weak_order[2])] += share
                else:
                    raise NotImplementedError
        return d

    @cached_property
    def d_ballot_share_weak_voters_sincere(self):
        """dict : Ballot shares due to the weak orders if they vote sincerely

        Voters of the type ``'a>b~c'``:

        * In Approval or Plurality, they vote for `a`.
        * In Anti-plurality, half of them vote for `ab` (i.e. against `c`) and half of them vote for `ac` (i.e.
          against `b`).

        Voters of the type ``'a~b>c'``:

        * In Approval or Anti-plurality, they vote for `ab` (i.e. against `c`).
        * In Plurality, half of them vote for `a` and half of them vote for `b`.
        """
        d = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for weak_order, share in self.d_weak_order_share.items():
            if is_lover(weak_order):
                if self.voting_rule in {APPROVAL, PLURALITY}:
                    d[weak_order[0]] += share
                elif self.voting_rule == ANTI_PLURALITY:
                    d[sort_ballot(weak_order[0] + weak_order[2])] += my_division(share, 2)
                    d[sort_ballot(weak_order[0] + weak_order[4])] += my_division(share, 2)
                else:
                    raise NotImplementedError
            else:  # is_hater(weak_order)
                if self.voting_rule == PLURALITY:
                    d[weak_order[0]] += my_division(share, 2)
                    d[weak_order[2]] += my_division(share, 2)
                elif self.voting_rule in {APPROVAL, ANTI_PLURALITY}:
                    d[sort_ballot(weak_order[0] + weak_order[2])] += share
                else:
                    raise NotImplementedError
        return d

    def is_equilibrium(self, strategy):
        """Whether a strategy is an equilibrium in this profile.

        Parameters
        ----------
        strategy : Strategy
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        EquilibriumStatus
            Whether `strategy` is an equilibrium in this profile.
        """
        raise NotImplementedError

    def best_responses_to_strategy(self, d_ranking_best_response):
        """Convert best responses to a :class:`StrategyThreshold`.

        Parameters
        ----------
        d_ranking_best_response : dict
            Key: ranking. Value: :class:`BestResponse`.

        Returns
        -------
        StrategyThreshold
            The conversion of the best responses into a strategy. Only the rankings present in this profile are
            mentioned in the strategy.
        """
        return StrategyThreshold({
            ranking: best_response.threshold_utility
            for ranking, best_response in d_ranking_best_response.items()
            if self.d_ranking_share[ranking] > 0
        }, ratio_optimistic=Fraction(1, 2), profile=self, voting_rule=self.voting_rule)

    @property
    def strategies_ordinal(self):
        """Iterator: ordinal strategies of the profile.

        Yields
        ------
        StrategyOrdinal
            All possible ordinal strategies for this profile.

        Examples
        --------
        Cf. :class:`ProfileOrdinal`.
        """
        return IterableStrategyOrdinal(profile=self)

    @property
    def strategies_pure(self):
        """Iterator: pure strategies of the profile.

        Yields
        ------
        Strategy
            All possible pure strategies of the profile. This is implemented only for discrete profiles such
            as :class:`ProfileTwelve` or :class:`ProfileDiscrete`.

        Examples
        --------
        Cf. :class:`ProfileDiscrete`.
        """
        raise NotImplementedError

    @property
    def strategies_group(self):
        """Iterator: group strategies of the profile.

        Yields
        ------
        Strategy
            All possible group strategies of the profile. This is implemented only for profiles where we consider
            that there is a natural notion of group, such as :class:`ProfileNoisyDiscrete`.

        Examples
        --------
        Cf. :class:`ProfileNoisyDiscrete`.
        """
        raise NotImplementedError

    def analyzed_strategies(self, strategies):
        """Analyze a list of strategies for the profile.

        Parameters
        ----------
        strategies : iterable
            An iterator of strategies, such as a list of strategies.

        Returns
        -------
        AnalyzedStrategies
            The analyzed strategies of the profile.

        Examples
        --------
            Cf. :meth:`ProfileOrdinal.analyzed_strategies_ordinal`.
        """
        equilibria = []
        utility_dependent = []
        inconclusive = []
        non_equilibria = []
        for s in strategies:
            strategy = s.deepcopy_with_attached_profile(profile=self)
            status = strategy.is_equilibrium
            if status == EquilibriumStatus.EQUILIBRIUM:
                equilibria.append(strategy)
            elif status == EquilibriumStatus.UTILITY_DEPENDENT:
                utility_dependent.append(strategy)
            elif status == EquilibriumStatus.INCONCLUSIVE:  # pragma: no cover
                inconclusive.append(strategy)
                raise AssertionError('Met an inconclusive case: \nprofile = %r\nstrategy = %r' % (self, strategy))
            else:
                non_equilibria.append(strategy)
        return AnalyzedStrategies(equilibria, utility_dependent, inconclusive, non_equilibria)

    @cached_property
    def analyzed_strategies_ordinal(self):
        """AnalyzedStrategies: Analyzed ordinal strategies.

        Cf. :meth:`analyzed_strategies` and :attr:`strategies_ordinal`.
        """
        return self.analyzed_strategies(self.strategies_ordinal)

    @cached_property
    def analyzed_strategies_pure(self):
        """AnalyzedStrategies: Analyzed pure strategies.

        Cf. :meth:`analyzed_strategies` and :attr:`strategies_pure`. This is implemented only for discrete profiles such
        as :class:`ProfileTwelve` or :class:`ProfileDiscrete`.
        """
        return self.analyzed_strategies(self.strategies_pure)

    @cached_property
    def analyzed_strategies_group(self):
        """AnalyzedStrategies: Analyzed group strategies.

        Cf. :meth:`analyzed_strategies` and :attr:`strategies_group`. This is implemented only for profiles where we
        consider that there is a natural notion of group, such as :class:`ProfileNoisyDiscrete`.
        """
        return self.analyzed_strategies(self.strategies_group)

    @classmethod
    def order_and_label(cls, t):
        """Order and label of a discrete type.

        Helper method for the ternary plots. Implemented only for subclasses of Profile where the initialization
        is made with a dictionary that maps types to voter share, such as :class:`ProfileNoisyDiscrete` or
        :class:`ProfileOrdinal`, but unlike :class:`ProfileHistogram` (where a second dictionary is needed for the
        histograms).

        Parameters
        ----------
        t : object
            A type. Any type that is accepted at a key in the initialization dictionary.

        Returns
        -------
        order : str
            The ranking or weak order.
        label : str
            The label to be used for the corner of the triangle.

        Examples
        --------
            Cf. :meth:`ProfileNoisyDiscrete.order_and_label`.
        """
        raise NotImplementedError

    @classmethod
    def order_and_label_weak(cls, t):
        r"""Auxiliary function for :meth:`order_and_label`, specialized for weak orders.

        Parameters
        ----------
        t : object
            A weak order of the form ``'a>b~c'`` or ``'a~b>c'``.

        Returns
        -------
        order : str
            The weak order itself.
        label : str
            The label to be used for the corner of the triangle.

        Examples
        --------
            >>> Profile.order_and_label_weak('a~b>c')
            ('a~b>c', '$r(a\\sim b>c)$')
        """
        return t, ('$r(%s)$' % t).replace('~', '\\sim ')

    def random_tau_undominated(self):
        """Random tau based on undominated ballots.

        This is used, for example, in :meth:`ProfileCardinal.iterated_voting`.

        Returns
        -------
        TauVector
            A random tau-vector. Independently for each ranking, a proportion uniformly drawn in [0, 1] of voters
            use one undominated ballot, and the rest use the other undominated ballot. For example, in Approval voting,
            voters with ranking `abc` are randomly split between ballots `a` and `ab`.
        """
        d = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for ranking, share in self.d_ranking_share.items():
            r = random.random()
            d[ballot_low_u(ranking, self.voting_rule)] += r * share
            d[ballot_high_u(ranking, self.voting_rule)] += (1 - r) * share
        return TauVector(d, voting_rule=self.voting_rule, symbolic=self.symbolic)


def make_property_ranking_share(ranking, doc):
    def _f(self):
        return self.d_ranking_share[ranking]
    _f.__doc__ = doc
    return property(_f)


for my_ranking in RANKINGS:
    setattr(Profile, my_ranking, make_property_ranking_share(my_ranking, 'Number : Share of voters with this ranking.'))



from poisson_approval.constants.constants import *
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.strategies.StrategyTwelve import StrategyTwelve
from poisson_approval.utils.Util import product_dict
from poisson_approval.utils.UtilBallots import ballot_low_u, ballot_high_u


class IterableStrategyTwelve:
    """Iterate over twelve-type strategies (:class:`StrategyTwelve`).

    Parameters
    ----------
    profile : Profile, optional
    voting_rule : str, optional
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``. Default: the same
        voting rule as `profile` if a profile is specified, ``APPROVAL`` otherwise.
    d_ranking_fixed_strategy : dict
        Key: ranking. Value: fixed strategy. Cf. examples below.
    test : callable.
        A function ``StrategyTwelve -> bool``. Only strategies meeting this test are given.

    Examples
    --------
    Basic usage:

        >>> for strategy in IterableStrategyTwelve():  # doctest: +ELLIPSIS
        ...     print(strategy)
        <abc: a, acb: a, bac: b, bca: b, cab: c, cba: c>
        <abc: a, acb: a, bac: b, bca: b, cab: c, cba: bc>
        <abc: a, acb: a, bac: b, bca: b, cab: c, cba: utility-dependent>
        ...
        <abc: utility-dependent, acb: utility-dependent, bac: utility-dependent, bca: utility-dependent, \
cab: utility-dependent, cba: utility-dependent>

    Specify a profile:

        >>> from poisson_approval import ProfileTwelve
        >>> profile = ProfileTwelve({'ab_c': 0.5, 'a_bc': 0.25, 'b_ac': 0.25})
        >>> for strategy in IterableStrategyTwelve(profile=profile):
        ...     print(strategy)
        <abc: a, bac: b> ==> a
        <abc: a, bac: ab> ==> a
        <abc: ab, bac: b> ==> b
        <abc: ab, bac: ab> ==> a, b
        <abc: utility-dependent, bac: b> ==> a, b
        <abc: utility-dependent, bac: ab> ==> a

    Specify some fixed strategies:

        >>> iterable = IterableStrategyTwelve(
        ...     d_ranking_fixed_strategy={'abc': 'a', 'acb': 'a', 'bac': 'b', 'bca': UTILITY_DEPENDENT})
        >>> for strategy in iterable:
        ...     print(strategy)
        <abc: a, acb: a, bac: b, bca: utility-dependent, cab: c, cba: c>
        <abc: a, acb: a, bac: b, bca: utility-dependent, cab: c, cba: bc>
        <abc: a, acb: a, bac: b, bca: utility-dependent, cab: c, cba: utility-dependent>
        <abc: a, acb: a, bac: b, bca: utility-dependent, cab: ac, cba: c>
        <abc: a, acb: a, bac: b, bca: utility-dependent, cab: ac, cba: bc>
        <abc: a, acb: a, bac: b, bca: utility-dependent, cab: ac, cba: utility-dependent>
        <abc: a, acb: a, bac: b, bca: utility-dependent, cab: utility-dependent, cba: c>
        <abc: a, acb: a, bac: b, bca: utility-dependent, cab: utility-dependent, cba: bc>
        <abc: a, acb: a, bac: b, bca: utility-dependent, cab: utility-dependent, cba: utility-dependent>

    Use a condition with the parameter `test`:

        >>> profile = ProfileTwelve({'ab_c': 0.5, 'a_bc': 0.25, 'b_ac': 0.25})
        >>> def test_a_wins(strategy):
        ...     return strategy.winners == {'a'}
        >>> for strategy in IterableStrategyTwelve(profile=profile, test=test_a_wins):
        ...     print(strategy)
        <abc: a, bac: b> ==> a
        <abc: a, bac: ab> ==> a
        <abc: utility-dependent, bac: ab> ==> a
    """

    def __init__(self, profile=None, voting_rule=None, d_ranking_fixed_strategy=None, test=None):
        # Default parameters
        if d_ranking_fixed_strategy is None:
            d_ranking_fixed_strategy = dict()
        # Parameters
        self.profile = profile
        # noinspection PyProtectedMember
        self.voting_rule = Strategy._get_voting_rule_(profile, voting_rule)
        self.d_ranking_fixed_strategy = d_ranking_fixed_strategy
        self.test = test
        # Computed variables

        def possible_ballots(ranking):
            try:
                return [self.d_ranking_fixed_strategy[ranking]]
            except KeyError:
                pass
            if self.profile is not None:
                share_ranking_1 = self.profile.d_type_share[ranking[0] + '_' + ranking[1:]]
                share_ranking_12 = self.profile.d_type_share[ranking[0:2] + '_' + ranking[2]]
                strategy_1 = ballot_low_u(ranking, self.voting_rule)
                strategy_12 = ballot_high_u(ranking, self.voting_rule)
                if share_ranking_1 > 0 and share_ranking_12 > 0:
                    return [strategy_1, strategy_12, UTILITY_DEPENDENT]
                elif share_ranking_1 > 0 or share_ranking_12 > 0:
                    return [strategy_1, strategy_12]
                else:
                    return ['']
            else:
                return [ballot_low_u(ranking, self.voting_rule), ballot_high_u(ranking, self.voting_rule),
                        UTILITY_DEPENDENT]

        self.d_ranking_possible_ballots = {ranking: possible_ballots(ranking) for ranking in RANKINGS}

    def __iter__(self):
        for d_ranking_ballot in product_dict(self.d_ranking_possible_ballots):
            strategy = StrategyTwelve(d_ranking_ballot, profile=self.profile, voting_rule=self.voting_rule)
            if self.test is None or self.test(strategy):
                yield strategy



import math
import numpy as np
import sympy as sp
from fractions import Fraction
from poisson_approval.utils.ComputationEngine import ComputationEngine


class ComputationEngineSymbolic(ComputationEngine):
    """Computation engine: symbolic computation.

    This engine relies on the external package `sympy` in order to perform symbolic computation:

        >>> ce = ComputationEngineSymbolic
        >>> ce.inf
        oo
        >>> ce.nan
        nan
        >>> ce.exp(3)
        exp(3)
        >>> ce.log(3)
        log(3)
        >>> ce.Rational(1, 3)
        1/3
        >>> ce.S(1) / 3
        1/3
        >>> ce.simplify(- ce.Rational(1, 10) - (- ce.sqrt(15) / 5 + ce.sqrt(30) / 10)**2)
        -1 + 3*sqrt(2)/5
        >>> ce.sqrt(3)
        sqrt(3)

    Usage of :meth:`look_equal`:

        >>> ce.look_equal(1, 0.999999999999)
        True
        >>> ce.look_equal(1, np.float(0.999999999999))
        True
        >>> ce.look_equal(1, sp.Float(0.999999999999))
        False
        >>> ce.look_equal(1, Fraction(999999999999, 1000000000000))
        False
        >>> ce.look_equal(ce.sqrt(2), ce.Rational(14142135623730951, 10000000000000000))
        False
    """

    # Constants

    inf = sp.oo
    nan = sp.nan
    pi = sp.pi

    # Functions

    @classmethod
    def exp(cls, x):
        return sp.exp(x)

    @classmethod
    def factorial(cls, x):
        return sp.factorial(x)

    @classmethod
    def log(cls, x):
        return sp.log(x)

    @classmethod
    def Rational(cls, x, y):
        return sp.Rational(x, y)

    @classmethod
    def S(cls, x):
        return sp.S(x)

    @classmethod
    def simplify(cls, x):
        return sp.simplify(x, ratio=1)

    @classmethod
    def sqrt(cls, x):
        return sp.sqrt(x)




import numpy as np
from scipy.optimize import minimize
from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event


class EventTrio(Event):
    """A 3-candidate tie.

    Notes
    -----
    We consider the 3-way tie, i.e. situations where ``S_x = S_y = S_z``.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> from poisson_approval import TauVector
        >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(6, 10), 'c': Fraction(3, 10)})
        >>> EventTrio(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
        >>> from poisson_approval import TauVector
        >>> tau = TauVector({'a': Fraction(1, 6), 'b': Fraction(1, 6), 'c': Fraction(1, 6),
        ...                  'ab': Fraction(1, 6), 'ac': Fraction(1, 6), 'bc': Fraction(1, 6)})
        >>> EventTrio(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        <asymptotic = exp(? log n + ? + o(1)), phi_a = 1, phi_b = 1, phi_c = 1, phi_ab = 1, phi_ac = 1, phi_bc = 1>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        ce = self.ce
        is_cross_diagram = (tau_x == 0 and tau_yz == 0) or (tau_y == 0 and tau_xz == 0) or (tau_z == 0 and tau_xy == 0)
        is_flower_diagram = ((tau_x == 0 and tau_xy == 0) or (tau_x == 0 and tau_xz == 0)
                             or (tau_y == 0 and tau_xy == 0) or (tau_y == 0 and tau_yz == 0)
                             or (tau_z == 0 and tau_xz == 0) or (tau_z == 0 and tau_yz == 0))
        if is_cross_diagram or is_flower_diagram:
            self.asymptotic = (Asymptotic.poisson_eq(tau_x, tau_yz, symbolic=self.symbolic)
                               * Asymptotic.poisson_eq(tau_y, tau_xz, symbolic=self.symbolic)
                               * Asymptotic.poisson_eq(tau_z, tau_xy, symbolic=self.symbolic))
            self._phi_x = ce.simplify(ce.sqrt(ce.S(tau_yz) / tau_x)) if tau_x > 0 else ce.nan
            self._phi_y = ce.simplify(ce.sqrt(ce.S(tau_xz) / tau_y)) if tau_y > 0 else ce.nan
            self._phi_z = ce.simplify(ce.sqrt(ce.S(tau_xy) / tau_z)) if tau_z > 0 else ce.nan
            self._phi_yz = ce.simplify(ce.sqrt(ce.S(tau_x) / tau_yz)) if tau_yz > 0 else ce.nan
            self._phi_xz = ce.simplify(ce.sqrt(ce.S(tau_y) / tau_xz)) if tau_xz > 0 else ce.nan
            self._phi_xy = ce.simplify(ce.sqrt(ce.S(tau_z) / tau_xy)) if tau_xy > 0 else ce.nan
        elif tau_xy == 0 and tau_xz == 0 and tau_yz == 0:
            # Tripod. Note that the other coefficient are not 0, otherwise it would be a (degenerate) cross diagram.
            self.asymptotic = Asymptotic(
                mu=ce.simplify(3 * ce.S(tau_x * tau_y * tau_z) ** ce.Rational(1, 3) - 1), nu=ce.nan, xi=ce.nan,
                symbolic=self.symbolic)
            self._phi_x = ce.simplify((ce.S(tau_y * tau_z) / tau_x ** 2) ** ce.Rational(1, 3))
            self._phi_y = ce.simplify((ce.S(tau_x * tau_z) / tau_y ** 2) ** ce.Rational(1, 3))
            self._phi_z = ce.simplify((ce.S(tau_x * tau_y) / tau_z ** 2) ** ce.Rational(1, 3))
            self._phi_xy = ce.nan
            self._phi_xz = ce.nan
            self._phi_yz = ce.nan
        elif tau_x == 0 and tau_y == 0 and tau_z == 0:
            # Inverted tripod. Note that the other coefficient are not 0, otherwise it would be a (degenerate) cross
            # diagram.
            self.asymptotic = Asymptotic(
                mu=ce.simplify(3 * ce.S(tau_xy * tau_xz * tau_yz) ** ce.Rational(1, 3) - 1), nu=ce.nan, xi=ce.nan,
                symbolic=self.symbolic)
            self._phi_x = ce.nan
            self._phi_y = ce.nan
            self._phi_z = ce.nan
            self._phi_xy = ce.simplify((ce.S(tau_xz * tau_yz) / tau_xy ** 2) ** ce.Rational(1, 3))
            self._phi_xz = ce.simplify((ce.S(tau_xy * tau_yz) / tau_xz ** 2) ** ce.Rational(1, 3))
            self._phi_yz = ce.simplify((ce.S(tau_xy * tau_xz) / tau_yz ** 2) ** ce.Rational(1, 3))
        elif tau_x - tau_yz == tau_y - tau_xz == tau_z - tau_xy:
            # Natural general tie.
            # This would work in the general case, but we can avoid numeric approximations easily!
            self.asymptotic = Asymptotic(mu=0, nu=ce.nan, xi=ce.nan, symbolic=self.symbolic)
            self._phi_x = ce.S(1) if tau_x > 0 else ce.nan
            self._phi_y = ce.S(1) if tau_y > 0 else ce.nan
            self._phi_z = ce.S(1) if tau_z > 0 else ce.nan
            self._phi_xy = ce.S(1) if tau_xy > 0 else ce.nan
            self._phi_xz = ce.S(1) if tau_xz > 0 else ce.nan
            self._phi_yz = ce.S(1) if tau_yz > 0 else ce.nan
        else:
            # Generic case: we work with floats, not sympy expressions.
            tau_x_f, tau_y_f, tau_z_f = float(tau_x), float(tau_y), float(tau_z)
            tau_xy_f, tau_xz_f, tau_yz_f = float(tau_xy), float(tau_xz), float(tau_yz)
            inf, sup, start = self._get_bounds_and_start(tau_x_f, tau_y_f, tau_z_f,
                                                         tau_xy_f, tau_xz_f, tau_yz_f)
            # Let's go for the actual computation
            optimizer = minimize(
                lambda x: 2 * np.sqrt(
                    (tau_x_f * x + tau_xz_f) * (tau_yz_f / x + tau_y_f)
                ) + tau_xy_f * x + tau_z_f / x - 1,
                start,
                bounds=[(inf, sup)]
            )
            self.asymptotic = Asymptotic(mu=ce.S(float(optimizer.fun)), nu=ce.nan, xi=ce.nan, symbolic=self.symbolic)
            x_2 = ce.S(optimizer.x[0])
            x_1 = ce.simplify(ce.sqrt((ce.S(tau_yz) / x_2 + tau_y) / (tau_x * x_2 + tau_xz)))
            self._phi_x = ce.simplify(x_1 * x_2) if tau_x > 0 else ce.nan
            self._phi_y = ce.simplify(ce.S(1) / x_1) if tau_y > 0 else ce.nan
            self._phi_xz = x_1 if tau_xz > 0 else ce.nan
            self._phi_yz = ce.simplify(ce.S(1) / (x_1 * x_2)) if tau_yz > 0 else ce.nan
            self._phi_xy = x_2 if tau_xy > 0 else ce.nan
            self._phi_z = ce.simplify(1 / x_2) if tau_z > 0 else ce.nan

    def _get_bounds_and_start(self, tau_x_f, tau_y_f, tau_z_f,
                              tau_xy_f, tau_xz_f, tau_yz_f):
        """Return inf, sup, start.

        >>> from poisson_approval import TauVector
        >>> tau = TauVector({'a': 1/3, 'ac': 1/3, 'b': 1/6, 'bc': 1/6})
        >>> event = EventTrio(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        >>> event._get_bounds_and_start(event._tau_x, event._tau_y, event._tau_z,
        ...                             event._tau_xy, event._tau_xz, event._tau_yz)
        (1, 1, 1)
        >>> event = EventTrio(candidate_x='b', candidate_y='c', candidate_z='a', tau=tau)
        >>> event._get_bounds_and_start(event._tau_x, event._tau_y, event._tau_z,
        ...                             event._tau_xy, event._tau_xz, event._tau_yz)
        (1.4142135623730951, 1.4142135623730951, 1.4142135623730951)
        >>> event = EventTrio(candidate_x='c', candidate_y='a', candidate_z='b', tau=tau)
        >>> event._get_bounds_and_start(event._tau_x, event._tau_y, event._tau_z,
        ...                             event._tau_xy, event._tau_xz, event._tau_yz)
        (0.7071067811865476, 0.7071067811865476, 0.7071067811865476)
        """

        # Use pivot xy
        score_xy_in_pivot_xy = getattr(self.tau, 'score_%s_in_duo_%s' % (self._label_xy, self._label_xy))
        score_z_in_pivot_xy = getattr(self.tau, 'score_%s_in_duo_%s' % (self._label_z, self._label_xy))
        if score_xy_in_pivot_xy > score_z_in_pivot_xy:
            # Easy pivot      => phi_z > 1 => x_2 < 1
            inf, sup = 0, 1 - 1e-14
        elif score_xy_in_pivot_xy < score_z_in_pivot_xy:
            # Difficult pivot => phi_z < 1 => x_2 > 1
            inf, sup = 1 + 1e-14, np.inf
        else:
            # Tight pivot     => x_2 = 1
            return 1, 1, 1

        # Use pivot xz
        if tau_x_f == 0 and tau_xz_f <= tau_y_f:
            pass
        else:
            if tau_x_f == 0:
                root = tau_yz_f / (tau_xz_f - tau_y_f)
            else:
                a = tau_x_f
                b = tau_xz_f - tau_y_f
                c = - tau_yz_f
                root = (- b + np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            score_xz_in_pivot_xz = getattr(self.tau, 'score_%s_in_duo_%s' % (self._label_xz, self._label_xz))
            score_y_in_pivot_xz = getattr(self.tau, 'score_%s_in_duo_%s' % (self._label_y, self._label_xz))
            if score_xz_in_pivot_xz > score_y_in_pivot_xz:
                # Easy pivot      => phi_y > 1 => x_1 < 1 => x_2 > root
                inf = max(inf, root + 1e-14)
            elif score_xz_in_pivot_xz < score_y_in_pivot_xz:
                # Difficult pivot => phi_y < 1 => x_1 > 1 => x_2 < root
                sup = min(sup, root - 1e-14)
            else:
                # Tight pivot     => x_1 = 1 => x_2 = root
                return root, root, root

        # Use pivot yz
        if tau_y_f == 0 and tau_yz_f <= tau_x_f:
            pass
        else:
            if tau_y_f == 0:
                root = tau_xz_f / (tau_yz_f - tau_x_f)
            else:
                a = tau_y_f
                b = tau_yz_f - tau_x_f
                c = - tau_xz_f
                root = (- b + np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            score_yz_in_pivot_yz = getattr(self.tau, 'score_%s_in_duo_%s' % (self._label_yz, self._label_yz))
            score_x_in_pivot_yz = getattr(self.tau, 'score_%s_in_duo_%s' % (self._label_x, self._label_yz))
            if score_yz_in_pivot_yz > score_x_in_pivot_yz:
                # Easy pivot      => phi_x > 1 => x_1 * x_2 > 1 => x_2 > root
                inf = max(inf, root + 1e-14)
            elif score_yz_in_pivot_yz < score_x_in_pivot_yz:
                # Difficult pivot => phi_x < 1 => x_1 * x_2 < 1 => x_2 < root
                sup = min(sup, root - 1e-14)
            else:
                # Tight pivot     => x_1 * x_2 = 1 => x_2 = root
                return root, root, root

        # Conclude
        assert inf <= sup
        if inf == 0 and np.isposinf(sup):  # pragma: no cover
            start = 1
        elif inf == 0:
            start = sup / 2
        elif np.isposinf(sup):
            start = inf * 2
        else:
            start = np.sqrt(inf * sup)
        if inf == 0:
            inf = 1e-14
        return inf, sup, start



from poisson_approval.constants.constants import *
from poisson_approval.random_factories.RandSimplexGridUniform import RandSimplexGridUniform
from poisson_approval.utils.Util import initialize_random_seeds, rand_simplex_grid
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal
from poisson_approval.profiles.ProfileHistogram import ProfileHistogram


class RandProfileHistogramGridUniform(RandSimplexGridUniform):
    """A random factory of histogram profiles (:class:`ProfileHistogram`), uniform on a grid.

    Parameters
    ----------
    denominator : int
        The grain of the grid for the shares of the rankings.
    denominator_bins : int
        The grain of the grid for the bins.
    n_bins : int
        The number of bins in each histogram. Cf. :class:`ProfileHistogram`.
    orders : list, optional
        These orders will have a variable share. They can be rankings, e.g. ``'abc'``, or weak orders, e.g.
        ``'a~b>c'``. Default: all rankings.
    d_order_fixed_share : dict, optional.
        A dictionary. For each entry ``order: fixed_share``, this order will have at least this fixed share. The total
        must be lower or equal to 1.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileHistogram`.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_profile = RandProfileHistogramGridUniform(denominator=17, denominator_bins=7, n_bins=3)
        >>> profile = rand_profile()
        >>> print(profile)
        <abc: 1/17 [Fraction(2, 7) 0 Fraction(5, 7)], \
acb: 8/17 [Fraction(1, 7) Fraction(2, 7) Fraction(4, 7)], \
bac: 2/17 [Fraction(2, 7) Fraction(1, 7) Fraction(4, 7)], \
cab: 5/17 [Fraction(1, 7) Fraction(6, 7) 0], \
cba: 1/17 [Fraction(3, 7) Fraction(4, 7) 0]> (Condorcet winner: a)

    Using the optional parameters:

        >>> from fractions import Fraction
        >>> rand_profile = RandProfileHistogramGridUniform(
        ...     denominator=17, denominator_bins=7,
        ...     n_bins=3, orders=['abc', 'a~b>c'], d_order_fixed_share={'b>a~c': Fraction(1, 2)},
        ...     voting_rule=PLURALITY)
        >>> profile = rand_profile()
        >>> print(profile)
        <abc: 1/2 [Fraction(6, 7) Fraction(1, 7) 0], b>a~c: 1/2> (Condorcet winner: a, b) (Plurality)
    """

    def __init__(self, denominator, denominator_bins, n_bins, orders=None, d_order_fixed_share=None, **kwargs):
        if orders is None:
            orders = RANKINGS
        self.denominator = denominator
        self.denominator_bins = denominator_bins
        self.n_bins = n_bins
        self.kwargs_histogram = kwargs
        super().__init__(cls=ProfileOrdinal, denominator=denominator, keys=orders,
                         d_key_fixed_share=d_order_fixed_share)

    def __call__(self):
        profile_ordinal = super().__call__()
        d_ranking_histogram = {}
        for ranking in sorted(profile_ordinal.support_in_rankings):
            d_ranking_histogram[ranking] = rand_simplex_grid(d=self.n_bins, denominator=self.denominator_bins)
        return ProfileHistogram(d_ranking_share=profile_ordinal.d_ranking_share,
                                d_ranking_histogram=d_ranking_histogram,
                                d_weak_order_share=profile_ordinal.d_weak_order_share,
                                **self.kwargs_histogram)



from fractions import Fraction
from poisson_approval import TauVector, BestResponseAntiPlurality, ANTI_PLURALITY, UTILITY_DEPENDENT


def test_best_response_anti_plurality():
    """
        >>> tau = TauVector({'ab': Fraction(9, 15), 'ac': Fraction(4, 15), 'bc': Fraction(2, 15)},
        ...                 voting_rule=ANTI_PLURALITY)
        >>> best_response = BestResponseAntiPlurality(tau, 'abc')
        >>> print(best_response._str_very_verbose)
        tau = <ab: 3/5, ac: 4/15, bc: 2/15> ==> a (Anti-plurality)
        ranking = abc
        voting_rule  = Anti-plurality
        duo_ij = <asymptotic = exp(- 0.0228764 n - 0.5 log n - 0.431347 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        duo_ij = <asymptotic = exp(- 0.0228764 n - 0.5 log n - 0.431347 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        duo_ji = <asymptotic = exp(- 0.0228764 n - 0.5 log n - 0.431347 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        duo_ik = <asymptotic = exp(- 0.167648 n - 0.5 log n - 0.63408 + o(1)), phi_ab = 0.471405, phi_ac = 1, phi_bc = 2.12132>
        duo_ki = <asymptotic = exp(- 0.167648 n - 0.5 log n - 0.63408 + o(1)), phi_ab = 0.471405, phi_ac = 1, phi_bc = 2.12132>
        duo_jk = <asymptotic = exp(- 0.0666667 n - 0.5 log n - 0.807367 + o(1)), phi_ab = 0.666667, phi_ac = 1.5, phi_bc = 1>
        duo_kj = <asymptotic = exp(- 0.0666667 n - 0.5 log n - 0.807367 + o(1)), phi_ab = 0.666667, phi_ac = 1.5, phi_bc = 1>
        pivot_weak_ij = <asymptotic = exp(- 0.0228764 n - 0.5 log n - 0.431347 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        pivot_weak_ji = <asymptotic = exp(- 0.0228764 n - 0.5 log n - 0.431347 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        pivot_weak_ik = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_weak_ki = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_weak_jk = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_weak_kj = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_strict_ij = <asymptotic = exp(- 0.0228764 n - 0.5 log n - 0.431347 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        pivot_strict_ji = <asymptotic = exp(- 0.0228764 n - 0.5 log n - 0.431347 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        pivot_strict_ik = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_strict_ki = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_strict_jk = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_strict_kj = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_tij_ijk = <asymptotic = exp(- 0.0228764 n - 0.5 log n + 0.103453 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        pivot_tij_ikj = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_tij_jik = <asymptotic = exp(- 0.0228764 n - 0.5 log n + 0.450026 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        pivot_tij_jki = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_tij_kij = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_tij_kji = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_tjk_ijk = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_tjk_ikj = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_tjk_jik = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_tjk_jki = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_tjk_kij = <asymptotic = exp(- 0.0228764 n - 0.5 log n + 0.103453 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        pivot_tjk_kji = <asymptotic = exp(- 0.0228764 n - 0.5 log n + 0.450026 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        pivot_tij = <asymptotic = exp(- 0.0228764 n - 0.5 log n + 0.103453 + o(1)), phi_ab = 1, phi_ac = 0.707107, phi_bc = 1.41421>
        pivot_tjk = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio_1t_i = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio_1t_j = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio_1t_k = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio_2t_ij = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio_2t_ji = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio_2t_ik = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio_2t_ki = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio_2t_jk = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        trio_2t_kj = <asymptotic = exp(- 0.167966 n + ? log n + ? + o(1)), phi_ab = 0.462241, phi_ac = 1.04004, phi_bc = 2.08008>
        pivot_ij_easy_or_tight = True
        pivot_ji_easy_or_tight = True
        pivot_ik_easy_or_tight = False
        pivot_ki_easy_or_tight = False
        pivot_jk_easy_or_tight = False
        pivot_kj_easy_or_tight = False
        threshold_utility = 1
        justification = Anti-plurality analysis
        ballot = ac
    """
    pass


def test_best_response_is_ij():
    tau = TauVector({'ab': Fraction(4, 15), 'ac': Fraction(9, 15), 'bc': Fraction(2, 15)}, voting_rule=ANTI_PLURALITY)
    best_response = BestResponseAntiPlurality(tau, 'abc')
    assert best_response.ballot == 'ab'


def test_best_response_is_utility_dependent():
    tau = TauVector({'ab': Fraction(7, 15), 'ac': Fraction(1, 15), 'bc': Fraction(7, 15)}, voting_rule=ANTI_PLURALITY)
    best_response = BestResponseAntiPlurality(tau, 'abc')
    assert best_response.threshold_utility == Fraction(1, 2)
    assert best_response.ballot == UTILITY_DEPENDENT



from poisson_approval.utils.ComputationEngineSymbolic import ComputationEngineSymbolic
from poisson_approval.utils.ComputationEngineNumeric import ComputationEngineNumeric


def computation_engine(symbolic):
    """Choose the computation engine.

    Parameters
    ----------
    symbolic : bool

    Returns
    -------
    ComputationEngine
        :class:`ComputationEngineSymbolic` if `symbolic` is True, :class:`ComputationEngineNumeric` otherwise.
    """
    return ComputationEngineSymbolic if symbolic else ComputationEngineNumeric



from poisson_approval.constants.constants import *
from poisson_approval.random_factories.RandSimplexUniform import RandSimplexUniform
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.utils.Util import initialize_random_seeds
from poisson_approval.utils.UtilBallots import allowed_ballots


class RandTauVectorUniform(RandSimplexUniform):
    """A random factory of tau-vectors (:class:`TauVector`) following the uniform distribution.

    Parameters
    ----------
    ballots : iterable, optional
        These ballots (e.g. ``'a'``, ``'ab'``) will have a variable share. Default: all allowed ballots.
    d_ballot_fixed_share : dict, optional
        A dictionary. For each entry ``ballot: fixed_share``, this ballot will have at least this fixed share. The total
        must be lower or equal to 1.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`TauVector`.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_tau = RandTauVectorUniform()
        >>> tau = rand_tau()
        >>> print(tau)
        <a: 0.4236547993389047, ab: 0.12122838365799216, ac: 0.0039303209304278885, b: 0.05394987214431912, \
bc: 0.1124259903007756, c: 0.2848106336275805> ==> a

    If the voting rule is not approval, only the relevant ballots are used:

        >>> initialize_random_seeds()
        >>> rand_tau = RandTauVectorUniform(voting_rule=PLURALITY)
        >>> tau = rand_tau()
        >>> print(tau)
        <a: 0.5488135039273248, b: 0.16637586244509472, c: 0.2848106336275805> ==> a (Plurality)

    Using the optional parameters:

        >>> rand_tau = RandTauVectorUniform(
        ...     ballots=['b', 'bc'], d_ballot_fixed_share={'a': 0.5},
        ...     symbolic=True)
        >>> tau = rand_tau()
        >>> print(tau)
        <a: 0.5, b: 0.30138168803582194, bc: 0.19861831196417806> ==> a, b

    For more examples, cf. :class:`RandSimplexUniform`.
    """

    def __init__(self, ballots=None, d_ballot_fixed_share=None, **kwargs):
        if ballots is None:
            try:
                ballots = allowed_ballots(kwargs['voting_rule'])
            except KeyError:
                ballots = allowed_ballots()
        super().__init__(cls=TauVector, keys=ballots, d_key_fixed_share=d_ballot_fixed_share, **kwargs)



import warnings
from poisson_approval.constants.constants import *
from poisson_approval.profiles.ProfileCardinalContinuous import ProfileCardinalContinuous
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.Util import product_dict, my_division
from poisson_approval.utils.UtilPreferences import is_weak_order, sort_weak_order, is_hater
from poisson_approval.utils.UtilCache import cached_property


class ProfileNoisyDiscrete(ProfileCardinalContinuous):
    """Profile with a discrete distribution of voters, with noise.

    Parameters
    ----------
    d : dict
        Cf. examples below for the different types of input syntax.
    noise : Number, optional
        Cf. examples below for the different types of input syntax.
    d_weak_order_share : dict
        E.g. ``{'a~b>c': 0.2, 'a>b~c': 0.1}``. ``d_weak_order_share['a~b>c']`` is the probability that a voter likes
        candidates ``a`` and ``b`` equally and prefer them to candidate ``c``.
    normalization_warning : bool
        Whether a warning should be issued if the input distribution is not normalized.
    ratio_sincere : Number
        The ratio of sincere voters, in the interval [0, 1]. This is used for :meth:`tau`.
    ratio_fanatic : Number
        The ratio of fanatic voters, in the interval [0, 1]. This is used for :meth:`tau`. The sum of `ratio_sincere`
        and `ratio_fanatic` must not exceed 1.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.
    symbolic : bool
        Whether the computations are symbolic or numeric.

    Attributes
    ----------
    d_ranking_utility_noise_share : dict of dict
        It maps a ranking to a dict that maps a tuple (utility, noise) to a share of voters. It means that this share
        of voters have this ranking and with a utility for their second candidate that is uniformly distributed in an
        interval [umin, umax], where umin = max(utility - noise, 0) and umax = min(utility + noise, 1).

    Notes
    -----
    If the input distribution is not normalized, the profile will be normalized anyway and a warning is
    issued (unless `normalization_warning` is False).

    Examples
    --------
    The four following examples illustrate different ways to define the same profile, where:

    * 26/100 of the voters have ranking `abc` and a utility for `b` that is uniformly distributed between 0.3-0.01 and
      0.3+0.01,
    * 53/100 of the voters have ranking `abc` and a utility for `b` that is uniformly distributed between 0.8-0.01 and
      0.8+0.01,
    * 21/100 of the voters have ranking `bac` and a utility for `a` that is uniformly distributed between 0.3-0.01 and
      0.1+0.01,

    The first possible type of input syntax is a dict that maps a tuple (ranking, utility) to a share of voters. All
    groups of voters have the same noise about their utility, given by the parameter `noise`:

        >>> from fractions import Fraction
        >>> profile = ProfileNoisyDiscrete({
        ...     ('abc', 0.3): Fraction(26, 100),
        ...     ('abc', 0.8): Fraction(53, 100),
        ...     ('bac', 0.1): Fraction(21, 100)
        ... }, noise=0.01)
        >>> print(profile)
        <abc 0.3 ± 0.01: 13/50, abc 0.8 ± 0.01: 53/100, bac 0.1 ± 0.01: 21/100> (Condorcet winner: a)

    The second type of input syntax has the advantage of not repeating a ranking if is it used for several groups of
    voters. It is a dict that maps a ranking to a nested dictionary, itself mapping a utility to a share of voters.
    Like above, all groups of voters have the same noise, given by the parameter `noise`:

        >>> from fractions import Fraction
        >>> profile = ProfileNoisyDiscrete({
        ...     'abc': {0.3: Fraction(26, 100), 0.8: Fraction(53, 100)},
        ...     'bac': {0.1: Fraction(21, 100)}
        ... }, noise=0.01)
        >>> print(profile)
        <abc 0.3 ± 0.01: 13/50, abc 0.8 ± 0.01: 53/100, bac 0.1 ± 0.01: 21/100> (Condorcet winner: a)

    The two remaining types of input syntax are variants of the two types presented above. They are a bit more verbose
    but they enable to specify a specific noise for each group of voters.

    Here is the variant of the first type of syntax. It is a dict that maps tuple (ranking, utility, noise) to
    a share of voters. It corresponds exactly to the attribute :attr:`d_ranking_utility_noise_share`:

        >>> from fractions import Fraction
        >>> profile = ProfileNoisyDiscrete({
        ...     ('abc', 0.3, 0.01): Fraction(26, 100),
        ...     ('abc', 0.8, 0.01): Fraction(53, 100),
        ...     ('bac', 0.1, 0.01): Fraction(21, 100)
        ... })
        >>> print(profile)
        <abc 0.3 ± 0.01: 13/50, abc 0.8 ± 0.01: 53/100, bac 0.1 ± 0.01: 21/100> (Condorcet winner: a)

    And here is the variant of the second type of syntax. It is a dict that maps a ranking to a nested dict, itself
    mapping a tuple (utility, noise) to a share of voters:

        >>> from fractions import Fraction
        >>> profile = ProfileNoisyDiscrete({
        ...     'abc': {(0.3, 0.01): Fraction(26, 100), (0.8, 0.01): Fraction(53, 100)},
        ...     'bac': {(0.1, 0.01): Fraction(21, 100)}
        ... })
        >>> print(profile)
        <abc 0.3 ± 0.01: 13/50, abc 0.8 ± 0.01: 53/100, bac 0.1 ± 0.01: 21/100> (Condorcet winner: a)

    Some examples of operations on the profile:

        >>> profile
        ProfileNoisyDiscrete({'abc': {(0.3, 0.01): Fraction(13, 50), (0.8, 0.01): Fraction(53, 100)}, \
'bac': {(0.1, 0.01): Fraction(21, 100)}})
        >>> profile.d_ranking_share
        {'abc': Fraction(79, 100), 'bac': Fraction(21, 100)}
        >>> profile.abc
        Fraction(79, 100)
        >>> profile.have_ranking_with_utility_above_u('abc', 0.5)
        Fraction(53, 100)
        >>> profile.have_ranking_with_utility_below_u('abc', 0.5)
        Fraction(13, 50)
        >>> profile.have_ranking_with_utility_u('abc', 0.3)
        0
        >>> profile.d_candidate_welfare
        {'a': 0.811, 'b': 0.712, 'c': 0}
        >>> profile.analyzed_strategies_group
        Equilibrium:
        <abc: a, bac: b> ==> a (FF)
        <BLANKLINE>
        Non-equilibria:
        <abc: ab, bac: ab> ==> a, b (FF)
        <abc: ab, bac: b> ==> b (FF)
        <abc: utility-dependent (0.55), bac: ab> ==> a (FF)
        <abc: utility-dependent (0.55), bac: b> ==> a (FF)
        <abc: a, bac: ab> ==> a (FF)
        >>> print(profile.analyzed_strategies_group.winners_at_equilibrium)
        a

    The profile can include weak orders:

        >>> profile = ProfileNoisyDiscrete(
        ...     {('abc', 0.3, 0.01): Fraction(26, 100), ('bac', 0.1, 0.01): Fraction(21, 100)},
        ...     d_weak_order_share={'a~b>c': Fraction(53, 100)})
        >>> profile
        ProfileNoisyDiscrete({'abc': {(0.3, 0.01): Fraction(13, 50)}, 'bac': {(0.1, 0.01): Fraction(21, 100)}}, \
d_weak_order_share={'a~b>c': Fraction(53, 100)})
        >>> print(profile)
        <abc 0.3 ± 0.01: 13/50, bac 0.1 ± 0.01: 21/100, a~b>c: 53/100> (Condorcet winner: a)

    Please note that the voters with a weak order are not "noisy".
    """

    def __init__(self, d, noise=None, d_weak_order_share=None, normalization_warning=True,
                 ratio_sincere=0, ratio_fanatic=0, voting_rule=APPROVAL, symbolic=False):
        super().__init__(ratio_sincere=ratio_sincere, ratio_fanatic=ratio_fanatic, voting_rule=voting_rule,
                         symbolic=symbolic)
        self.d_ranking_utility_noise_share = DictPrintingInOrderIgnoringZeros({
            ranking: DictPrintingInOrderIgnoringZeros() for ranking in RANKINGS})
        if d_weak_order_share is None:
            d_weak_order_share = dict()
        self._d_weak_order_share = DictPrintingInOrderIgnoringZeros({
            weak_order: 0 for weak_order in WEAK_ORDERS_WITHOUT_INVERSIONS})
        # Input d

        def add_voters(r, u, epsilon, s):
            # Ranking r, utility u, share s, noise epsilon
            if s == 0:
                return
            elif epsilon == 0:  # pragma: no cover
                raise ValueError('Noise should be > 0')
            else:
                self.d_ranking_utility_noise_share[r][(u, epsilon)] = (
                    self.d_ranking_utility_noise_share[r].get((u, epsilon), 0) + s)

        for key, value in d.items():
            if is_weak_order(key):
                self._d_weak_order_share[sort_weak_order(key)] += value
            elif isinstance(key, tuple):
                if len(key) == 2:
                    add_voters(*key, noise, value)
                else:  # len(tuple) == 3
                    add_voters(*key, value)
            else:  # isinstance(key, str)
                d_key2_share = value
                for key2, share in d_key2_share.items():
                    if isinstance(key2, tuple):
                        add_voters(key, *key2, share)
                    else:  # key2 is a number
                        add_voters(key, key2, noise, share)
        # Input d_weak_order_share
        for weak_order, share in d_weak_order_share.items():
            self._d_weak_order_share[sort_weak_order(weak_order)] += share
        # Normalize if necessary
        total = (sum([sum(d_utility_noise_share.values())
                      for d_utility_noise_share in self.d_ranking_utility_noise_share.values()])
                 + sum(self._d_weak_order_share.values()))
        if not self.ce.look_equal(total, 1):
            if normalization_warning:
                warnings.warn(NORMALIZATION_WARNING)
            for d_utility_noise_share in self.d_ranking_utility_noise_share.values():
                for utility_noise, share in d_utility_noise_share.items():
                    d_utility_noise_share[utility_noise] = my_division(share, total)
            for weak_order in self._d_weak_order_share.keys():
                self._d_weak_order_share[weak_order] = my_division(self._d_weak_order_share[weak_order], total)

    @cached_property
    def d_ranking_share(self):
        return DictPrintingInOrderIgnoringZeros({
            ranking: sum(d_utility_noise_share.values())
            for ranking, d_utility_noise_share in self.d_ranking_utility_noise_share.items()
        })

    @cached_property
    def d_ranking_umin_umax_share(self):
        """dict: Dictionary that maps a ranking to a dictionary that maps a tuple (u_min, u_max) to a share of voters.
        """
        return {ranking: {(max(utility - noise, 0), min(utility + noise, 1)): share
                          for (utility, noise), share in d_utility_noise_share.items()}
                for ranking, d_utility_noise_share in self.d_ranking_utility_noise_share.items()}

    @cached_property
    def d_weak_order_share(self):
        return self._d_weak_order_share

    def have_ranking_with_utility_above_u(self, ranking, u):
        d_umin_umax_share = self.d_ranking_umin_umax_share[ranking]
        return sum([
            _crop(my_division(umax - u, umax - umin)) * share
            for (umin, umax), share in d_umin_umax_share.items()
        ])

    def have_ranking_with_utility_below_u(self, ranking, u):
        d_umin_umax_share = self.d_ranking_umin_umax_share[ranking]
        return sum([
            _crop(my_division(u - umin, umax - umin)) * share
            for (umin, umax), share in d_umin_umax_share.items()
        ])

    def __repr__(self):
        """
            >>> from fractions import Fraction
            >>> profile = ProfileNoisyDiscrete({
            ...     ('abc', 0.3, 0.01): Fraction(26, 100),
            ...     ('abc', 0.8, 0.01): Fraction(53, 100),
            ...     ('bac', 0.1, 0.01): Fraction(21, 100)
            ... }, ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(2, 10))
            >>> profile
            ProfileNoisyDiscrete({'abc': {(0.3, 0.01): Fraction(13, 50), (0.8, 0.01): Fraction(53, 100)}, \
'bac': {(0.1, 0.01): Fraction(21, 100)}}, ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(1, 5))
        """
        arguments = repr(self.d_ranking_utility_noise_share)
        if self.contains_weak_orders:
            arguments += ', d_weak_order_share=%r' % self.d_weak_order_share
        if self.ratio_sincere > 0:
            arguments += ', ratio_sincere=%r' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            arguments += ', ratio_fanatic=%r' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'ProfileNoisyDiscrete(%s)' % arguments

    def __str__(self):
        """
            >>> from fractions import Fraction
            >>> profile = ProfileNoisyDiscrete({
            ...     ('abc', 0.3, 0.01): Fraction(26, 100),
            ...     ('abc', 0.8, 0.01): Fraction(53, 100),
            ...     ('bac', 0.1, 0.01): Fraction(21, 100)
            ... }, ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(2, 10))
            >>> print(profile)
            <abc 0.3 ± 0.01: 13/50, abc 0.8 ± 0.01: 53/100, bac 0.1 ± 0.01: 21/100> (Condorcet winner: a) \
(ratio_sincere: 1/10) (ratio_fanatic: 1/5)
        """
        contents = []
        if self.contains_rankings:
            contents.append(', '.join([
                '%s %s ± %s: %s' % (ranking, utility, noise,
                                    self.d_ranking_utility_noise_share[ranking][(utility, noise)])
                for ranking in sorted(self.d_ranking_utility_noise_share.keys())
                if self.d_ranking_utility_noise_share[ranking]
                for utility, noise in sorted(self.d_ranking_utility_noise_share[ranking].keys())
            ]))
        if self.contains_weak_orders:
            contents.append(str(self.d_weak_order_share)[1:-1])
        result = '<' + ', '.join(contents) + '>'
        if self.is_profile_condorcet:
            result += ' (Condorcet winner: %s)' % self.condorcet_winners
        if self.ratio_sincere > 0:
            result += ' (ratio_sincere: %s)' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            result += ' (ratio_fanatic: %s)' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            result += ' (%s)' % self.voting_rule
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : Object

        Returns
        -------
        bool
            True iff this profile is equal to `other`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileNoisyDiscrete({
            ...     ('abc', 0.3, 0.01): Fraction(26, 100),
            ...     ('abc', 0.8, 0.01): Fraction(53, 100),
            ...     ('bac', 0.1, 0.01): Fraction(21, 100)
            ... })
            >>> profile == ProfileNoisyDiscrete({
            ...     ('abc', 0.3, 0.01): Fraction(26, 100),
            ...     ('abc', 0.8, 0.01): Fraction(53, 100),
            ...     ('bac', 0.1, 0.01): Fraction(21, 100)
            ... })
            True
        """
        return (isinstance(other, ProfileNoisyDiscrete)
                and all([self.d_ranking_utility_noise_share[ranking] == other.d_ranking_utility_noise_share[ranking]
                         for ranking in RANKINGS])
                and self.d_weak_order_share == other.d_weak_order_share
                and self.ratio_sincere == other.ratio_sincere
                and self.ratio_fanatic == other.ratio_fanatic
                and self.voting_rule == other.voting_rule)

    @cached_property
    def standardized_version(self):
        """ProfileNoisyDiscrete : Standardized version of the profile (makes it unique, up to permutations of the
        candidates).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileNoisyDiscrete({
            ...     ('abc', 0.3, 0.01): Fraction(26, 100),
            ...     ('abc', 0.8, 0.01): Fraction(53, 100),
            ...     ('bac', 0.1, 0.01): Fraction(21, 100)
            ... })
            >>> print(profile.standardized_version)
            <abc 0.3 ± 0.01: 13/50, abc 0.8 ± 0.01: 53/100, bac 0.1 ± 0.01: 21/100> (Condorcet winner: a)
            >>> profile.is_standardized
            True
        """
        def translate(s, permute):
            return s.replace('a', permute[0]).replace('b', permute[1]).replace('c', permute[2])

        best_d = {}
        best_signature = []
        for perm in XYZ_PERMUTATIONS:
            d_test = {translate(ranking, perm): d_utility_noise_share
                      for ranking, d_utility_noise_share in self.d_ranking_utility_noise_share.items()}
            d_test.update({sort_weak_order(translate(weak_order, perm)): share
                           for weak_order, share in self.d_weak_order_share.items()})
            signature_test = [[(utility, d_test[ranking][utility]) for utility in sorted(d_test[ranking].keys())]
                              for ranking in XYZ_RANKINGS]
            signature_test += [d_test[weak_order] for weak_order in XYZ_WEAK_ORDERS_WITHOUT_INVERSIONS]
            if signature_test > best_signature:
                best_signature = signature_test
                best_d = d_test
        return ProfileNoisyDiscrete(
            {ranking: best_d[xyz_ranking] for ranking, xyz_ranking in zip(RANKINGS, XYZ_RANKINGS)},
            {weak_order: best_d[xyz_weak_order] for weak_order, xyz_weak_order in zip(
                WEAK_ORDERS_WITHOUT_INVERSIONS, XYZ_WEAK_ORDERS_WITHOUT_INVERSIONS)},
            ratio_sincere=self.ratio_sincere, ratio_fanatic=self.ratio_fanatic,
            voting_rule=self.voting_rule)

    @cached_property
    def d_candidate_welfare(self):
        d = {candidate: 0 for candidate in CANDIDATES}
        for ranking, d_umin_umax_share in self.d_ranking_umin_umax_share.items():
            for (u_min, u_max), share in d_umin_umax_share.items():
                d[ranking[0]] += share
                d[ranking[1]] += share * (u_min + u_max) / 2
        for weak_order, share in self.d_weak_order_share.items():
            if share > 0:
                d[weak_order[0]] += share
                if is_hater(weak_order):
                    d[weak_order[2]] += share
        return d

    @property
    def strategies_group(self):
        """Iterator: group strategies of the profile.

        Yields
        ------
        StrategyThreshold
            All possible group strategies of the profile.
        """
        def possible_thresholds(ranking):
            if self.d_ranking_share[ranking] == 0:
                return [None]
            d_utility_noise_share = self.d_ranking_utility_noise_share[ranking]
            utilities = sorted([utility for utility, noise in d_utility_noise_share.keys()])
            return [0] + [my_division(x + y, 2) for x, y in zip(utilities[:-1], utilities[1:])] + [1]

        d_ranking_possible_thresholds = {ranking: possible_thresholds(ranking) for ranking in RANKINGS}

        for d_ranking_threshold in product_dict(d_ranking_possible_thresholds):
            yield StrategyThreshold(d_ranking_threshold, profile=self)

    @property
    def strategies_pure(self):
        raise NotImplementedError

    @classmethod
    def order_and_label(cls, t):
        r"""Order and label of a discrete type.

        Cf. :meth:`Profile.order_and_label`.

        Examples
        --------
            >>> ProfileNoisyDiscrete.order_and_label(('abc', 0.5, 0.01))
            ('abc', '$r(abc, u_b = 0.5 ± 0.01)$')
            >>> ProfileNoisyDiscrete.order_and_label(('abc', 0.5))
            ('abc', '$r(abc, u_b \\simeq 0.5)$')
            >>> ProfileNoisyDiscrete.order_and_label('a~b>c')
            ('a~b>c', '$r(a\\sim b>c)$')
        """
        if isinstance(t, tuple):
            if len(t) == 3:
                return t[0], '$r(%s, u_%s = %s ± %s)$' % (t[0], t[0][1], t[1], t[2])
            else:  # len(t) == 2
                return t[0], '$r(%s, u_%s \\simeq %s)$' % (t[0], t[0][1], t[1])
        else:
            return cls.order_and_label_weak(t)


def _crop(x, low=0, high=1):
    """Crop a number to an interval.

    Parameters
    ----------
    x : Number
    low : Number
    high : Number

    Returns
    -------
    Number
        `x` in the general case, but `low` if `x` is lower and `high` if `x` is higher.

    Examples
    --------
        >>> _crop(- 0.2)
        0
        >>> _crop(0.5)
        0.5
        >>> _crop(1.5)
        1
    """
    return min(max(x, low), high)



from itertools import product
from poisson_approval.constants.constants import *
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.utils.Util import my_division


class IterableStrategyThresholdGrid:
    """Iterate over threshold strategies (:class:`StrategyThreshold`).

    Parameters
    ----------
    denominator_threshold : int or iterable
        The grain(s) of the grid for the utility thresholds.
    denominator_ratio_optimistic : int or iterable, optional
        The grain(s) of the grid for the ratios of optimistic voters.
    profile : Profile, optional
    voting_rule : str, optional
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``. Default: the same
        voting rule as `profile` if a profile is specified, ``APPROVAL`` otherwise.
    d_ranking_fixed_strategy : dict
        Key: ranking. Value: fixed strategy. Cf. examples below.
    test : callable.
        A function ``StrategyThreshold -> bool``. Only strategies meeting this test are given.

    Examples
    --------
    Basic usage:

        >>> for strategy in IterableStrategyThresholdGrid(denominator_threshold=3):  # doctest: +ELLIPSIS
        ...     print(strategy)
        <abc: ab, acb: ac, bac: ab, bca: bc, cab: ac, cba: bc>
        <abc: ab, acb: ac, bac: ab, bca: bc, cab: ac, cba: utility-dependent (1/3)>
        <abc: ab, acb: ac, bac: ab, bca: bc, cab: ac, cba: utility-dependent (2/3)>
        <abc: ab, acb: ac, bac: ab, bca: bc, cab: ac, cba: c>
        <abc: ab, acb: ac, bac: ab, bca: bc, cab: utility-dependent (1/3), cba: bc>
        ...
        <abc: a, acb: a, bac: b, bca: b, cab: c, cba: c>

    Specify a profile:

        >>> from poisson_approval import ProfileHistogram
        >>> profile = ProfileHistogram({'abc': 0.75, 'bac': 0.25}, {'abc': [1], 'bac': [1]})
        >>> for strategy in IterableStrategyThresholdGrid(denominator_threshold=2, profile=profile):
        ...     print(strategy)
        <abc: ab, bac: ab> ==> a, b
        <abc: ab, bac: utility-dependent (1/2)> ==> b
        <abc: ab, bac: b> ==> b
        <abc: utility-dependent (1/2), bac: ab> ==> a
        <abc: utility-dependent (1/2), bac: utility-dependent (1/2)> ==> a
        <abc: utility-dependent (1/2), bac: b> ==> a
        <abc: a, bac: ab> ==> a
        <abc: a, bac: utility-dependent (1/2)> ==> a
        <abc: a, bac: b> ==> a

    Specify some fixed strategies:

        >>> iterable = IterableStrategyThresholdGrid(
        ...     denominator_threshold=2,
        ...     d_ranking_fixed_strategy={'abc': 1, 'acb': 1, 'bac': 1, 'bca': 1})
        >>> for strategy in iterable:
        ...     print(strategy)
        <abc: a, acb: a, bac: b, bca: b, cab: ac, cba: bc>
        <abc: a, acb: a, bac: b, bca: b, cab: ac, cba: utility-dependent (1/2)>
        <abc: a, acb: a, bac: b, bca: b, cab: ac, cba: c>
        <abc: a, acb: a, bac: b, bca: b, cab: utility-dependent (1/2), cba: bc>
        <abc: a, acb: a, bac: b, bca: b, cab: utility-dependent (1/2), cba: utility-dependent (1/2)>
        <abc: a, acb: a, bac: b, bca: b, cab: utility-dependent (1/2), cba: c>
        <abc: a, acb: a, bac: b, bca: b, cab: c, cba: bc>
        <abc: a, acb: a, bac: b, bca: b, cab: c, cba: utility-dependent (1/2)>
        <abc: a, acb: a, bac: b, bca: b, cab: c, cba: c>

    Use a condition with the parameter `test`:

        >>> profile = ProfileHistogram({'abc': 0.75, 'bac': 0.25}, {'abc': [1], 'bac': [1]})
        >>> def test_a_wins(strategy):
        ...     return strategy.winners == {'a'}
        >>> for strategy in IterableStrategyThresholdGrid(denominator_threshold=2, profile=profile, test=test_a_wins):
        ...     print(strategy)
        <abc: utility-dependent (1/2), bac: ab> ==> a
        <abc: utility-dependent (1/2), bac: utility-dependent (1/2)> ==> a
        <abc: utility-dependent (1/2), bac: b> ==> a
        <abc: a, bac: ab> ==> a
        <abc: a, bac: utility-dependent (1/2)> ==> a
        <abc: a, bac: b> ==> a

    Loop also over the ratios of optimistic voters:

        >>> from poisson_approval import ProfileDiscrete
        >>> profile = ProfileDiscrete({('abc', 0.2): 0.75, ('bac', 0.4): 0.25})
        >>> iterable = IterableStrategyThresholdGrid(denominator_threshold=2, denominator_ratio_optimistic=2,
        ...                                          profile=profile)
        >>> for strategy in iterable:  # doctest: +ELLIPSIS
        ...     print(repr(strategy))
        StrategyThreshold({'abc': 0, 'bac': 0})
        StrategyThreshold({'abc': 0, 'bac': (Fraction(1, 2), 0)})
        StrategyThreshold({'abc': 0, 'bac': (Fraction(1, 2), Fraction(1, 2))})
        StrategyThreshold({'abc': 0, 'bac': (Fraction(1, 2), 1)})
        StrategyThreshold({'abc': 0, 'bac': 1})
        StrategyThreshold({'abc': (Fraction(1, 2), 0), 'bac': 0})
        ...
        StrategyThreshold({'abc': 1, 'bac': 1})
    """

    def __init__(self, denominator_threshold, denominator_ratio_optimistic=None, profile=None, voting_rule=None,
                 d_ranking_fixed_strategy=None, test=None, **kwargs):
        """
        We can use an iterable for `denominator_threshold` and `denominator_ratio_optimistic`:

            >>> from poisson_approval import ProfileDiscrete
            >>> profile = ProfileDiscrete({('abc', 0.2): 0.75})
            >>> iterable = IterableStrategyThresholdGrid(denominator_threshold=[2, 3],
            ...                                          denominator_ratio_optimistic=[2],
            ...                                          profile=profile)
            >>> for strategy in iterable:
            ...     print(repr(strategy))
            StrategyThreshold({'abc': 0})
            StrategyThreshold({'abc': (Fraction(1, 2), 0)})
            StrategyThreshold({'abc': (Fraction(1, 2), Fraction(1, 2))})
            StrategyThreshold({'abc': (Fraction(1, 2), 1)})
            StrategyThreshold({'abc': 1})
            StrategyThreshold({'abc': 0})
            StrategyThreshold({'abc': (Fraction(1, 3), 0)})
            StrategyThreshold({'abc': (Fraction(1, 3), Fraction(1, 2))})
            StrategyThreshold({'abc': (Fraction(1, 3), 1)})
            StrategyThreshold({'abc': (Fraction(2, 3), 0)})
            StrategyThreshold({'abc': (Fraction(2, 3), Fraction(1, 2))})
            StrategyThreshold({'abc': (Fraction(2, 3), 1)})
            StrategyThreshold({'abc': 1})
        """
        # Default parameters
        if d_ranking_fixed_strategy is None:
            d_ranking_fixed_strategy = dict()
        # Parameters
        self.denominator_threshold = denominator_threshold
        self.denominator_ratio_optimistic = denominator_ratio_optimistic
        self.profile = profile
        # noinspection PyProtectedMember
        self.voting_rule = Strategy._get_voting_rule_(profile, voting_rule)
        self.d_ranking_fixed_strategy = d_ranking_fixed_strategy
        self.test = test
        self.kwargs = kwargs
        # Computed variables
        if isinstance(self.denominator_threshold, int):
            self.denominators_threshold = [self.denominator_threshold]
        else:
            self.denominators_threshold = self.denominator_threshold
        if isinstance(self.denominator_ratio_optimistic, int) or self.denominator_ratio_optimistic is None:
            self.denominators_ratio = [self.denominator_ratio_optimistic]
        else:
            self.denominators_ratio = self.denominator_ratio_optimistic
        self.rankings_to_decide = RANKINGS if self.profile is None else self.profile.support_in_rankings
        self.rankings_to_decide = sorted(self.rankings_to_decide - self.d_ranking_fixed_strategy.keys())
        self.n_rankings_to_decide = len(self.rankings_to_decide)

    def __iter__(self):
        for denominator_t in self.denominators_threshold:
            for denominator_r in self.denominators_ratio:
                for tuple_thresholds in product(range(denominator_t + 1), repeat=self.n_rankings_to_decide):
                    if denominator_r is None:
                        d = {ranking: (my_division(threshold, denominator_t))
                             for ranking, threshold
                             in zip(self.rankings_to_decide, tuple_thresholds)}
                        d.update(self.d_ranking_fixed_strategy)
                        strategy = StrategyThreshold(d, profile=self.profile, voting_rule=self.voting_rule,
                                                     **self.kwargs)
                        if self.test is None or self.test(strategy):
                            yield strategy
                    else:
                        iterables_ratios = []
                        for threshold in tuple_thresholds:
                            if threshold in {0, denominator_t}:
                                iterables_ratios.append([None])
                            else:
                                iterables_ratios.append(range(denominator_r + 1))
                        for tuple_ratios in product(*iterables_ratios):
                            d = {ranking: (my_division(threshold, denominator_t),
                                           None if ratio is None else my_division(ratio, denominator_r))
                                 for ranking, threshold, ratio
                                 in zip(self.rankings_to_decide, tuple_thresholds, tuple_ratios)}
                            d.update(self.d_ranking_fixed_strategy)
                            strategy = StrategyThreshold(d, profile=self.profile, voting_rule=self.voting_rule,
                                                         **self.kwargs)
                            if self.test is None or self.test(strategy):
                                yield strategy



from poisson_approval.constants.constants import *
from poisson_approval.random_factories.RandSimplexGridUniform import RandSimplexGridUniform
from poisson_approval.utils.Util import initialize_random_seeds
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal


class RandProfileOrdinalGridUniform(RandSimplexGridUniform):
    """A random factory of ordinal profiles (:class:`ProfileOrdinal`), uniform on a grid

    Parameters
    ----------
    denominator : int
        The grain of the grid.
    orders : list, optional
        These orders will have a variable share. They can be rankings, e.g. ``'abc'``, or weak orders, e.g.
        ``'a~b>c'``. Default: all rankings.
    d_order_fixed_share : dict, optional.
        A dictionary. For each entry ``order: fixed_share``, this order will have at least this fixed share. The total
        must be lower or equal to 1.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileOrdinal`.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_profile = RandProfileOrdinalGridUniform(denominator=7)
        >>> profile = rand_profile()
        >>> print(profile)
        <abc: 2/7, acb: 1/7, bac: 1/7, bca: 3/7> (Condorcet winner: b)

    Using the optional parameters:

        >>> from fractions import Fraction
        >>> rand_profile = RandProfileOrdinalGridUniform(
        ...     denominator=4,
        ...     orders=['abc', 'acb'], d_order_fixed_share={'acb': Fraction(2, 7), 'a~b>c': Fraction(1, 7)},
        ...     voting_rule=PLURALITY)
        >>> profile = rand_profile()
        >>> print(profile)
        <abc: 3/7, acb: 3/7, a~b>c: 1/7> (Condorcet winner: a) (Plurality)

    For more examples, cf. :class:`RandSimplexGridUniform`.
    """

    def __init__(self, denominator, orders=None, d_order_fixed_share=None, **kwargs):
        if orders is None:
            orders = RANKINGS
        super().__init__(cls=ProfileOrdinal, denominator=denominator, keys=orders,
                         d_key_fixed_share=d_order_fixed_share, **kwargs)



import numpy as np
from poisson_approval.constants.constants import *
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.utils.Util import initialize_random_seeds, my_division


class RandStrategyThresholdGridUniform:
    """A random factory of threshold strategies (:class:`StrategyThreshold`), uniform on a grid.

    Parameters
    ----------
    denominator_threshold : int
        The grain of the grid for the utility thresholds.
    denominator_ratio_optimistic : int, optional
        The grain of the grid for the ratios of optimistic voters.
    profile : Profile, optional
    voting_rule : str, optional
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``. Default: the same
        voting rule as `profile` if a profile is specified, ``APPROVAL`` otherwise.
    d_ranking_fixed_strategy : dict
        Key: ranking. Value: fixed strategy. Cf. examples below.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_strategy = RandStrategyThresholdGridUniform(denominator_threshold=7)
        >>> strategy = rand_strategy()
        >>> print(strategy)
        <abc: utility-dependent (4/7), acb: a, bac: utility-dependent (5/7), bca: bc, cab: utility-dependent (3/7), \
cba: utility-dependent (3/7)>

    Specify a profile:

        >>> from poisson_approval import ProfileHistogram
        >>> profile = ProfileHistogram({'abc': 0.75, 'bac': 0.25}, {'abc': [1], 'bac': [1]})
        >>> rand_strategy = RandStrategyThresholdGridUniform(denominator_threshold=7, profile=profile)
        >>> strategy = rand_strategy()
        >>> print(strategy)
        <abc: utility-dependent (3/7), bac: b> ==> a

    Specify some fixed strategies:

        >>> from fractions import Fraction
        >>> rand_strategy = RandStrategyThresholdGridUniform(
        ...     denominator_threshold=7, denominator_ratio_optimistic=17,
        ...     d_ranking_fixed_strategy={'abc': 1, 'acb': 1, 'bac': 1, 'bca': (Fraction(1, 2), Fraction(1, 2))})
        >>> print(rand_strategy())
        <abc: a, acb: a, bac: b, bca: utility-dependent (1/2, 1/2), cab: utility-dependent (1/7, 4/17), cba: c>
    """

    def __init__(self, denominator_threshold, denominator_ratio_optimistic=None, profile=None, voting_rule=None,
                 d_ranking_fixed_strategy=None):
        # Default parameters
        if d_ranking_fixed_strategy is None:
            d_ranking_fixed_strategy = dict()
        # Parameters
        self.denominator_threshold = denominator_threshold
        self.denominator_ratio_optimistic = denominator_ratio_optimistic
        self.profile = profile
        # noinspection PyProtectedMember
        self.voting_rule = Strategy._get_voting_rule_(profile, voting_rule)
        self.d_ranking_fixed_strategy = d_ranking_fixed_strategy

        # Computed variables
        self.rankings_to_decide = RANKINGS if self.profile is None else self.profile.support_in_rankings
        self.rankings_to_decide = sorted(self.rankings_to_decide - self.d_ranking_fixed_strategy.keys())
        self.n_rankings_to_decide = len(self.rankings_to_decide)

    def __call__(self):
        if self.denominator_ratio_optimistic is None:
            d = {ranking: my_division(np.random.randint(0, self.denominator_threshold + 1),
                                      self.denominator_threshold)
                 for ranking in self.rankings_to_decide}
        else:
            d = {ranking: (my_division(np.random.randint(0, self.denominator_threshold + 1),
                                       self.denominator_threshold),
                           my_division(np.random.randint(0, self.denominator_ratio_optimistic + 1),
                                       self.denominator_ratio_optimistic))
                 for ranking in self.rankings_to_decide}
        d.update(self.d_ranking_fixed_strategy)
        return StrategyThreshold(d, profile=self.profile, voting_rule=self.voting_rule)



from fractions import Fraction
from poisson_approval import ProfileOrdinal, StrategyOrdinal, PLURALITY, ANTI_PLURALITY, initialize_random_seeds


def test_normalization():
    profile = ProfileOrdinal(d_ranking_share={'abc': 1, 'acb': 1})
    assert profile.abc == 0.5


def test_plurality():
    """
        >>> profile = ProfileOrdinal(d_ranking_share={'abc': 1, 'acb': 1}, voting_rule=PLURALITY)
        >>> profile
        ProfileOrdinal({'abc': Fraction(1, 2), 'acb': Fraction(1, 2)}, voting_rule='Plurality')
        >>> print(profile)
        <abc: 1/2, acb: 1/2> (Condorcet winner: a) (Plurality)
        >>> profile.tau_fanatic
        TauVector({'a': Fraction(1, 1)}, voting_rule='Plurality')
    """
    pass


def test_anti_plurality():
    """
        >>> profile = ProfileOrdinal(d_ranking_share={'abc': 1, 'acb': 1}, voting_rule=ANTI_PLURALITY)
        >>> profile
        ProfileOrdinal({'abc': Fraction(1, 2), 'acb': Fraction(1, 2)}, voting_rule='Anti-plurality')
        >>> print(profile)
        <abc: 1/2, acb: 1/2> (Condorcet winner: a) (Anti-plurality)
        >>> profile.tau_fanatic
        TauVector({'ab': Fraction(1, 2), 'ac': Fraction(1, 2)}, voting_rule='Anti-plurality')
    """
    pass


def test_d_ballot_share_weak_voters_fanatic():
    """
        >>> profile = ProfileOrdinal({'a~b>c': Fraction(7, 10), 'c>a~b': Fraction(3, 10)})
        >>> profile.d_ballot_share_weak_voters_fanatic
        {'a': Fraction(7, 20), 'b': Fraction(7, 20), 'c': Fraction(3, 10), 'ab': 0, 'ac': 0, 'bc': 0}
        >>> profile.voting_rule = PLURALITY
        >>> profile.d_ballot_share_weak_voters_fanatic
        {'a': Fraction(7, 20), 'b': Fraction(7, 20), 'c': Fraction(3, 10), 'ab': 0, 'ac': 0, 'bc': 0}
        >>> profile.voting_rule = ANTI_PLURALITY
        >>> profile.d_ballot_share_weak_voters_fanatic
        {'a': 0, 'b': 0, 'c': 0, 'ab': Fraction(7, 10), 'ac': Fraction(3, 20), 'bc': Fraction(3, 20)}
    """
    pass


def test_d_ballot_share_weak_voters_sincere():
    """
        >>> profile = ProfileOrdinal({'a~b>c': Fraction(7, 10), 'c>a~b': Fraction(3, 10)})
        >>> profile.d_ballot_share_weak_voters_sincere
        {'a': 0, 'b': 0, 'c': Fraction(3, 10), 'ab': Fraction(7, 10), 'ac': 0, 'bc': 0}
        >>> profile.voting_rule = PLURALITY
        >>> profile.d_ballot_share_weak_voters_sincere
        {'a': Fraction(7, 20), 'b': Fraction(7, 20), 'c': Fraction(3, 10), 'ab': 0, 'ac': 0, 'bc': 0}
        >>> profile.voting_rule = ANTI_PLURALITY
        >>> profile.d_ballot_share_weak_voters_sincere
        {'a': 0, 'b': 0, 'c': 0, 'ab': Fraction(7, 10), 'ac': Fraction(3, 20), 'bc': Fraction(3, 20)}
    """
    pass


def test_no_equilibrium():
    """
        >>> profile = ProfileOrdinal({'abc': Fraction(2, 5), 'bca': Fraction(2, 5), 'cab': Fraction(1, 5)})
        >>> profile.analyzed_strategies_ordinal.winners_at_equilibrium
        Winners()
    """
    pass


def test_strategy_weak_order():
    """
        >>> profile = ProfileOrdinal({'abc': Fraction(1, 7), 'a~b>c': Fraction(2, 7), 'c>a~b': Fraction(4, 7)})
        >>> strategy = StrategyOrdinal({'abc': 'a'}, profile=profile)
        >>> print(strategy.tau)
        <a: 1/7, ab: 2/7, c: 4/7> ==> c
    """
    pass



from fractions import Fraction
from poisson_approval import TauVector


def test_limit_pivot():
    """
        >>> tau = TauVector({'a': Fraction(1, 3), 'b': Fraction(1, 3), 'c': Fraction(1, 3)})
        >>> tau.pivot_weak_ab.asymptotic
        Asymptotic(mu=0, nu=-0.5, xi=-1.4093531597105358)
    """
    pass



import numpy as np


WHITE = np.array([1., 1., 1.])
BRIGHTNESS = 0.5


def abc_to_rgb(abc):
    """Conversion from candidates to RGB color

    Parameters
    ----------
    abc : tuple, list or ndarray
        Array of size 3. Each coefficient is associated to a candidate a, b, c. Typically the coefficients sum to 1
        or less than 1.

    Returns
    -------
    ndarray
        Array of size 3. RGB color. Roughly speaking, candidate `a` is associated to red, `b` to green and `c` to blue,
        but we try to mix colors more reasonably than just an average in RGB space.

    Examples
    --------
        >>> abc_to_rgb([1, 0, 0])
        array([1. , 0.5, 0.5])
        >>> abc_to_rgb([.5, .5, 0])
        array([1. , 1. , 0.5])
        >>> abc_to_rgb([0., 0., 0])
        array([0.5, 0.5, 0.5])
    """
    norm_inf = np.max(abc)
    if norm_inf == 0:
        rgb = np.zeros(3)
    else:
        norm_1 = np.sum(abc)
        rgb = np.array(abc) * np.min([norm_1, 1]) / norm_inf
    return rgb * (1 - BRIGHTNESS) + WHITE * BRIGHTNESS


COLOR_NO_ONE = abc_to_rgb([0, 0, 0])
COLOR_A = abc_to_rgb([1, 0, 0])
COLOR_B = abc_to_rgb([0, 1, 0])
COLOR_C = abc_to_rgb([0, 0, 1])
COLOR_AB = abc_to_rgb([1, 1, 0])
COLOR_AC = abc_to_rgb([1, 0, 1])
COLOR_BC = abc_to_rgb([0, 1, 1])
COLOR_ABC = abc_to_rgb([1, 1, 1])


def _uni_mix(color_1, color_2, weight_1, weight_2):
    """Mix two colors, with a fade that emphasizes the two base colors.

    Parameters
    ----------
    color_1, color_2 : tuple, list or ndarray
        Arrays of size 3 in format RGB.
    weight_1, weight_2 : Number
        Respective weights of the two colors.

    Returns
    -------
    ndarray
        Array of size 3 in format RGB. The mix is not linear. Currently, the resulting color is a barycenter
        with respective weights ``weight_1**n`` and ``weight_2**n``, with ``n = 2.5``. This makes a "sharper"
        transition, so that we see more of the base colors and less of the mixes.

    Examples
    --------
        >>> import numpy as np
        >>> _uni_mix(color_1=[1, 0, 0], color_2=[0, 1, 0], weight_1=1, weight_2=0)
        array([1., 0., 0.])
        >>> _uni_mix(color_1=[1, 0, 0], color_2=[0, 1, 0], weight_1=.5, weight_2=.5)
        array([0.5, 0.5, 0. ])
        >>> _uni_mix(color_1=[1, 0, 0], color_2=[0, 1, 0], weight_1=0, weight_2=1)
        array([0., 1., 0.])
    """
    n = 2.5
    return (np.array(color_1) * weight_1 ** n + np.array(color_2) * weight_2 ** n) / (weight_1 ** n + weight_2 ** n)


def _palette(x, y):
    """Compute the data for a nice color palette.

    Parameters
    ----------
    x, y : Number
        Cartesian coordinates of the point.

    Returns
    -------
    ndarray
        Array of size 3 in format RGB. The resulting figure consists of:

        * A nice color palette in the unit disk,
        * A black circle on the unit circle,
        * White background outside.

    Examples
    --------
        >>> _palette(x=0, y=1.5)
        array([1., 1., 1.])
        >>> _palette(x=0, y=0.99)
        array([0., 0., 0.])
        >>> color_almost_a = _palette(x=0, y=0.95)
        >>> np.all(np.isclose(color_almost_a, [1, .5, .5], atol=1E-2))
        True
    """
    z = x + 1j * y
    r = np.abs(z)
    if r > 1:
        return np.array([1., 1., 1.])
    if r > 0.96:
        return np.array([0., 0., 0.])
    theta = (np.angle(z, 'deg') + 30) % 360 - 30
    for theta_sup, color_inf, color_sup in zip([30, 90, 150, 210, 270, 330],
                                               [COLOR_C, COLOR_AC, COLOR_A, COLOR_AB, COLOR_B, COLOR_BC],
                                               [COLOR_AC, COLOR_A, COLOR_AB, COLOR_B, COLOR_BC, COLOR_C]):
        if theta <= theta_sup:
            weight_inf = (theta_sup - theta) / 60
            color = _uni_mix(color_sup, color_inf, 1 - weight_inf, weight_inf)
            return _uni_mix(color, COLOR_ABC, r ** .8, 1 - r ** .8)


DATA_PALETTE = np.array([[_palette(x / 100, y / 100) for x in range(-100, 100)] for y in range(-100, 100)])
